"""
ResultInfo
==========
"""
import traceback
import warnings

from enum import Enum, unique
from types import SimpleNamespace
from ansys.dpf.gate import (
    result_info_capi,
    result_info_grpcapi,
    integral_types,
    label_space_capi,
    label_space_grpcapi,
    object_handler,
    data_processing_grpcapi,
    data_processing_capi,
)

from ansys.dpf.core import collection
from ansys.dpf.core import server as server_module
from ansys.dpf.core import available_result, support
from ansys.dpf.core.cyclic_support import CyclicSupport
from ansys.dpf.core.label_space import LabelSpace
from ansys.dpf.core.check_version import version_requires


@unique
class physics_types(Enum):
    """
    ``'Physics_types'`` enumerates the different types of physics that an analysis can have.
    """

    mechanical = 0
    thermal = 1
    magnetic = 2
    electric = 3
    unknown_physics = 4
    fluid = 5


@unique
class analysis_types(Enum):
    """``'Analysis_types'`` enumerates the different types of analysis."""

    static = 0
    buckling = 1
    modal = 2
    harmonic = 3
    cms = 4
    transient = 5
    msup = 6
    substruct = 7
    spectrum = 8
    unknown_analysis = 9


class ResultInfo:
    """Represents the result information.

    This class describes the metadata of the analysis and the available results.

    Parameters
    ----------
    result_info : ctypes.c_void_p, ansys.grpc.dpf.result_info_pb2.ResultInfo message

     server : ansys.dpf.core.server, optional
        Server with the channel connected to the remote or local instance.
        The default is ``None``, in which case an attempt is made to use the
        global server.

    Examples
    --------
    Explore the result info from the model

    >>> from ansys.dpf import core as dpf
    >>> from ansys.dpf.core import examples
    >>> transient = examples.download_transient_result()
    >>> model = dpf.Model(transient)
    >>> result_info = model.metadata.result_info # printable result_info

    >>> result_info.available_results[0].name
    'displacement'
    >>> result_info.available_results[0].homogeneity
    'length'

    """

    def __init__(self, result_info, server=None):
        """Initialize with a ResultInfo message"""
        # ############################
        # step 1: get server
        self._server = server_module.get_or_create_server(server)

        # step 2: get api
        self._api = self._server.get_api_for_type(
            capi=result_info_capi.ResultInfoCAPI,
            grpcapi=result_info_grpcapi.ResultInfoGRPCAPI,
        )

        # step3: init environment
        self._api.init_result_info_environment(self)  # creates stub when gRPC

        # step4: if object exists, take the instance, else create it
        if result_info is not None:
            if isinstance(result_info, ResultInfo):
                self._internal_obj = result_info._internal_obj
            else:
                self._internal_obj = result_info
        elif result_info is None:
            raise Exception("Result_info given is None")

    def __str__(self):
        try:
            txt = (
                "%s analysis\n" % self.analysis_type.capitalize()
                + "Unit system: %s\n" % self.unit_system
                + "Physics Type: %s\n" % self.physics_type.capitalize()
                + "Available results:\n"
            )
            for res in self.available_results:
                line = [
                    "",
                    "-",
                    f"{res.name}: {res.native_location} {res.physical_name}",
                ]
                txt += "{0:^4} {1:^2} {2:<30}".format(*line) + "\n"

            if self._server.meet_version("7.0"):
                qualifiers_labels = self.available_qualifier_labels
                if len(qualifiers_labels) > 0:
                    txt += "Available qualifier labels:\n"
                    for label in qualifiers_labels:
                        label_support = self.qualifier_label_support(label)
                        names_field = label_support.string_field_support_by_property("names")
                        label_names = names_field.data_as_list
                        label_values = names_field.scoping.ids.tolist()
                        txt += f"     - {label}: "
                        for i, value in enumerate(label_values):
                            label_values[i] = label_names[i] + f" ({value})"
                        txt += f"{', '.join(label_values)}\n"

            return txt
        except Exception as e:
            raise e
            from ansys.dpf.core.core import _description

            return _description(self._internal_obj, self._server)

    @property
    def _names(self):
        return [item.name for item in self.available_results]

    def __contains__(self, value):
        return value in self._names

    @property
    def analysis_type(self):
        """Retrieves the analysis type.

        Returns
        -------
        analysis_type : str
            Type of the analysis, such as static or transient.

        Examples
        --------
        >>> from ansys.dpf import core as dpf
        >>> from ansys.dpf.core import examples
        >>> transient = examples.download_transient_result()
        >>> model = dpf.Model(transient)
        >>> result_info = model.metadata.result_info
        >>> result_info.analysis_type
        'static'

        """
        return self._api.result_info_get_analysis_type_name(self)

    @property
    def physics_type(self):
        """Type of the physics.

        Examples
        --------
        Mechanical result

        >>> from ansys.dpf import core as dpf
        >>> from ansys.dpf.core import examples
        >>> transient = examples.download_transient_result()
        >>> model = dpf.Model(transient)
        >>> result_info = model.metadata.result_info
        >>> result_info.physics_type
        'mechanical'

        """
        return self._get_physics_type()

    def _get_physics_type(self):
        """
        Returns
        -------
        physics_type : str
            Type of the physics, such as mechanical or electric.
        """
        return self._api.result_info_get_physics_type_name(self)

    @property
    def n_results(self):
        """Number of results."""
        return self._api.result_info_get_number_of_results(self)

    @property
    def unit_system(self):
        """Unit system of the result."""
        return self._api.result_info_get_unit_system_name(self)

    @property
    def cyclic_symmetry_type(self):
        """Cyclic symmetry type of the result.

        Return
        ------
        cyclic_symmetry_type : str
            Cyclic symmetry type of the results. Options are ``"single_stage"``,
            ``"multi_stage"``, and ``"not_cyclic"``.
        """
        return self._api.result_info_get_cyclic_symmetry_type(self)

    @property
    def has_cyclic(self):
        """Check the result file for cyclic symmetry.

        Return
        ------
        has_cyclic : bool
            Returns ``True`` if the result file has cyclic symmetry or is multistage.
        """
        return self._api.result_info_has_cyclic_symmetry(self)

    @property
    def cyclic_support(self):
        """Cyclic expansion information for a result file that has cyclic symmetry or is multistage.

        Return
        ------
        cyclic_support : CyclicSupport

        Examples
        --------
        Get a cyclic support from a model.

        >>> from ansys.dpf.core import Model
        >>> from ansys.dpf.core import examples
        >>> multi_stage = examples.download_multi_stage_cyclic_result()
        >>> model = Model(multi_stage)
        >>> result_info = model.metadata.result_info
        >>> cyc_support = result_info.cyclic_support

        """
        if self._api.result_info_has_cyclic_symmetry(self):
            cyclic_support = self._api.result_info_get_cyclic_support(self)
            return CyclicSupport(cyclic_support=cyclic_support, server=self._server)

    @property
    def unit_system_name(self):
        """Name of the unit system."""
        return self._api.result_info_get_unit_system_name(self)

    @property
    def solver_version(self):
        """Version of the solver."""
        major = integral_types.MutableInt32()
        minor = integral_types.MutableInt32()
        _ = self._api.result_info_get_solver_version(self, major, minor)
        return str(int(major)) + "." + str(int(minor))

    @property
    def solver_date(self):
        """Date of the solver."""
        date = integral_types.MutableInt32()
        time = integral_types.MutableInt32()
        self._api.result_info_get_solve_date_and_time(self, date, time)
        return int(date)

    @property
    def solver_time(self):
        """Time of the solver."""
        date = integral_types.MutableInt32()
        time = integral_types.MutableInt32()
        self._api.result_info_get_solve_date_and_time(self, date, time)
        return int(time)

    @property
    def user_name(self):
        """Name of the user."""
        return self._api.result_info_get_user_name(self)

    @property
    def job_name(self):
        """Name of the job."""
        return self._api.result_info_get_job_name(self)

    @property
    def product_name(self):
        """Name of the product."""
        return self._api.result_info_get_product_name(self)

    @property
    def main_title(self):
        """Main title."""
        return self._api.result_info_get_main_title(self)

    @property
    def available_results(self):
        """Available results, containing all information about results
        present in the result files.

        Returns
        -------
        available_result : list[AvailableResult]
        """
        out = []
        for i in range(len(self)):
            out.append(self._get_result(i))
        return out

    @property
    def _data_processing_core_api(self):
        core_api = self._server.get_api_for_type(
            capi=data_processing_capi.DataProcessingCAPI,
            grpcapi=data_processing_grpcapi.DataProcessingGRPCAPI,
        )
        core_api.init_data_processing_environment(self)
        return core_api

    def _get_result(self, numres):
        """
        Parameters
        ----------
        numres : int
            Index of the requested result.

        Returns
        -------
        result : Result
        """
        if numres >= len(self):
            raise IndexError("There are only %d results" % len(self))
        elif numres < 0:
            raise IndexError("Result index must be greater than 0")

        name = self._api.result_info_get_result_name(self, numres)
        physic_name = self._api.result_info_get_result_physics_name(self, numres)
        dimensionality = self._api.result_info_get_result_dimensionality_nature(self, numres)
        n_comp = self._api.result_info_get_result_number_of_components(self, numres)
        unit_symbol = self._api.result_info_get_result_unit_symbol(self, numres)
        homogeneity = self._api.result_info_get_result_homogeneity(self, numres)
        try:
            loc_name = integral_types.MutableString(256)
            self._api.result_info_get_result_location(self, numres, loc_name)
            loc_name = str(loc_name)
        except AttributeError:
            if name in available_result._result_properties:
                loc_name = available_result._result_properties[name]["location"]
            else:
                loc_name = ""
        try:
            scripting_name = self._api.result_info_get_result_scripting_name(self, numres)
        except AttributeError:
            if name in available_result._result_properties:
                scripting_name = available_result._result_properties[name]["scripting_name"]
            else:
                scripting_name = available_result._remove_spaces(physic_name)
        num_sub_res = self._api.result_info_get_number_of_sub_results(self, numres)
        sub_res = {}
        for ires in range(num_sub_res):
            sub_res_name = self._api.result_info_get_sub_result_name(self, numres, ires)
            ssub_res_rec_name = integral_types.MutableString(256)
            self._api.result_info_get_sub_result_operator_name(
                self, numres, ires, ssub_res_rec_name
            )
            ssub_res_rec_name = str(ssub_res_rec_name)
            descr = self._api.result_info_get_sub_result_description(self, numres, ires)
            sub_res[sub_res_name] = [ssub_res_rec_name, descr]

        qualifiers = []
        qualifier_labels = {}
        if self._server.meet_version("5.0"):
            qual_obj = object_handler.ObjHandler(
                data_processing_api=self._data_processing_core_api,
                internal_obj=self._api.result_info_get_qualifiers_for_result(self, numres),
            )
            label_space_api = self._server.get_api_for_type(
                capi=label_space_capi.LabelSpaceCAPI,
                grpcapi=label_space_grpcapi.LabelSpaceGRPCAPI,
            )
            num_qual_obj = label_space_api.list_label_spaces_size(qual_obj)
            for ires in range(num_qual_obj):
                label_space = LabelSpace(
                    label_space=label_space_api.list_label_spaces_at(qual_obj, ires),
                    obj=self,
                    server=self._server,
                )
                qualifiers.append(label_space)
                label_space_dict = label_space.__dict__()
                for key in label_space_dict.keys():
                    value = label_space_dict[key]
                    label_support = self.qualifier_label_support(key)
                    names_field = label_support.string_field_support_by_property("names")
                    label_value = names_field.data_as_list[
                        names_field.scoping.ids.tolist().index(value)
                    ]
                    label_value = label_value + f" ({value})"
                    if key not in qualifier_labels.keys():
                        qualifier_labels[key] = [label_value]
                    if label_value not in qualifier_labels[key]:
                        qualifier_labels[key].append(label_value)

        availableresult = SimpleNamespace(
            name=name,
            physicsname=physic_name,
            ncomp=n_comp,
            dimensionality=dimensionality,
            homogeneity=homogeneity,
            unit=unit_symbol,
            sub_res=sub_res,
            properties={"loc_name": loc_name, "scripting_name": scripting_name},
            qualifiers=qualifiers,
            qualifier_labels=qualifier_labels,
        )
        return available_result.AvailableResult(availableresult)

    @property
    @version_requires("5.0")
    def available_qualifier_labels(self):
        """Returns a list of labels defining result qualifiers

        Returns
        -------
        list[str]

        Notes
        -----
        Available with server's version starting at 5.0.
        """
        coll_obj = collection.StringCollection(
            collection=self._api.result_info_get_available_qualifier_labels_as_string_coll(self),
            server=self._server,
        )
        return coll_obj.get_integral_entries()

    @version_requires("5.0")
    def qualifier_label_support(self, label):
        """Returns what supports an available qualifier label.

        Parameters
        ----------
        label: str

        Returns
        -------
        Support

        Notes
        -----
        Available with server's version starting at 5.0.
        """
        return support.Support(
            support=self._api.result_info_get_qualifier_label_support(self, label),
            server=self._server,
        )

    def __len__(self):
        try:
            return self.n_results
        except Exception as e:
            return 0

    def __iter__(self):
        for i in range(len(self)):
            yield self[i]

    def __getitem__(self, key):
        if isinstance(key, int):
            index = key
        elif isinstance(key, str):
            if key not in self._names:
                raise ValueError('Invalid key "%s"' % key)
            index = self._names.index(key)
        else:
            raise TypeError('"%s" is an invalid keytype' % type(key))

        return self._get_result(index)

    def __del__(self):
        try:
            self._deleter_func[0](self._deleter_func[1](self))
        except:
            warnings.warn(traceback.format_exc())



"""
decimate_mesh
=============
Autogenerated DPF operator classes.
"""
from warnings import warn
from ansys.dpf.core.dpf_operator import Operator
from ansys.dpf.core.inputs import Input, _Inputs
from ansys.dpf.core.outputs import Output, _Outputs
from ansys.dpf.core.operators.specification import PinSpecification, Specification


class decimate_mesh(Operator):
    """Decimate a meshed region

    Parameters
    ----------
    mesh : MeshedRegion
        Mesh to decimate
    preservation_ratio : float, optional
        Target ratio of elements to preserve, the
        actual number of elements preserved
        might differ. default value is 0.5.
    aggressiveness : int, optional
        Quality measure for the resulting decimated
        mesh. lower aggresiveness will
        provide a higher quality mesh with
        the tradeoff of higher execution
        time. value range is 0 to 150,
        default is 0.


    Examples
    --------
    >>> from ansys.dpf import core as dpf

    >>> # Instantiate operator
    >>> op = dpf.operators.mesh.decimate_mesh()

    >>> # Make input connections
    >>> my_mesh = dpf.MeshedRegion()
    >>> op.inputs.mesh.connect(my_mesh)
    >>> my_preservation_ratio = float()
    >>> op.inputs.preservation_ratio.connect(my_preservation_ratio)
    >>> my_aggressiveness = int()
    >>> op.inputs.aggressiveness.connect(my_aggressiveness)

    >>> # Instantiate operator and connect inputs in one line
    >>> op = dpf.operators.mesh.decimate_mesh(
    ...     mesh=my_mesh,
    ...     preservation_ratio=my_preservation_ratio,
    ...     aggressiveness=my_aggressiveness,
    ... )

    >>> # Get output data
    >>> result_mesh = op.outputs.mesh()
    """

    def __init__(
        self,
        mesh=None,
        preservation_ratio=None,
        aggressiveness=None,
        config=None,
        server=None,
    ):
        super().__init__(name="decimate_mesh", config=config, server=server)
        self._inputs = InputsDecimateMesh(self)
        self._outputs = OutputsDecimateMesh(self)
        if mesh is not None:
            self.inputs.mesh.connect(mesh)
        if preservation_ratio is not None:
            self.inputs.preservation_ratio.connect(preservation_ratio)
        if aggressiveness is not None:
            self.inputs.aggressiveness.connect(aggressiveness)

    @staticmethod
    def _spec():
        description = """Decimate a meshed region"""
        spec = Specification(
            description=description,
            map_input_pin_spec={
                0: PinSpecification(
                    name="mesh",
                    type_names=["abstract_meshed_region"],
                    optional=False,
                    document="""Mesh to decimate""",
                ),
                1: PinSpecification(
                    name="preservation_ratio",
                    type_names=["double"],
                    optional=True,
                    document="""Target ratio of elements to preserve, the
        actual number of elements preserved
        might differ. default value is 0.5.""",
                ),
                2: PinSpecification(
                    name="aggressiveness",
                    type_names=["int32"],
                    optional=True,
                    document="""Quality measure for the resulting decimated
        mesh. lower aggresiveness will
        provide a higher quality mesh with
        the tradeoff of higher execution
        time. value range is 0 to 150,
        default is 0.""",
                ),
            },
            map_output_pin_spec={
                0: PinSpecification(
                    name="mesh",
                    type_names=["abstract_meshed_region"],
                    optional=False,
                    document="""Decimated mesh with triangle elements""",
                ),
            },
        )
        return spec

    @staticmethod
    def default_config(server=None):
        """Returns the default config of the operator.

        This config can then be changed to the user needs and be used to
        instantiate the operator. The Configuration allows to customize
        how the operation will be processed by the operator.

        Parameters
        ----------
        server : server.DPFServer, optional
            Server with channel connected to the remote or local instance. When
            ``None``, attempts to use the global server.
        """
        return Operator.default_config(name="decimate_mesh", server=server)

    @property
    def inputs(self):
        """Enables to connect inputs to the operator

        Returns
        --------
        inputs : InputsDecimateMesh
        """
        return super().inputs

    @property
    def outputs(self):
        """Enables to get outputs of the operator by evaluating it

        Returns
        --------
        outputs : OutputsDecimateMesh
        """
        return super().outputs


class InputsDecimateMesh(_Inputs):
    """Intermediate class used to connect user inputs to
    decimate_mesh operator.

    Examples
    --------
    >>> from ansys.dpf import core as dpf
    >>> op = dpf.operators.mesh.decimate_mesh()
    >>> my_mesh = dpf.MeshedRegion()
    >>> op.inputs.mesh.connect(my_mesh)
    >>> my_preservation_ratio = float()
    >>> op.inputs.preservation_ratio.connect(my_preservation_ratio)
    >>> my_aggressiveness = int()
    >>> op.inputs.aggressiveness.connect(my_aggressiveness)
    """

    def __init__(self, op: Operator):
        super().__init__(decimate_mesh._spec().inputs, op)
        self._mesh = Input(decimate_mesh._spec().input_pin(0), 0, op, -1)
        self._inputs.append(self._mesh)
        self._preservation_ratio = Input(decimate_mesh._spec().input_pin(1), 1, op, -1)
        self._inputs.append(self._preservation_ratio)
        self._aggressiveness = Input(decimate_mesh._spec().input_pin(2), 2, op, -1)
        self._inputs.append(self._aggressiveness)

    @property
    def mesh(self):
        """Allows to connect mesh input to the operator.

        Mesh to decimate

        Parameters
        ----------
        my_mesh : MeshedRegion

        Examples
        --------
        >>> from ansys.dpf import core as dpf
        >>> op = dpf.operators.mesh.decimate_mesh()
        >>> op.inputs.mesh.connect(my_mesh)
        >>> # or
        >>> op.inputs.mesh(my_mesh)
        """
        return self._mesh

    @property
    def preservation_ratio(self):
        """Allows to connect preservation_ratio input to the operator.

        Target ratio of elements to preserve, the
        actual number of elements preserved
        might differ. default value is 0.5.

        Parameters
        ----------
        my_preservation_ratio : float

        Examples
        --------
        >>> from ansys.dpf import core as dpf
        >>> op = dpf.operators.mesh.decimate_mesh()
        >>> op.inputs.preservation_ratio.connect(my_preservation_ratio)
        >>> # or
        >>> op.inputs.preservation_ratio(my_preservation_ratio)
        """
        return self._preservation_ratio

    @property
    def aggressiveness(self):
        """Allows to connect aggressiveness input to the operator.

        Quality measure for the resulting decimated
        mesh. lower aggresiveness will
        provide a higher quality mesh with
        the tradeoff of higher execution
        time. value range is 0 to 150,
        default is 0.

        Parameters
        ----------
        my_aggressiveness : int

        Examples
        --------
        >>> from ansys.dpf import core as dpf
        >>> op = dpf.operators.mesh.decimate_mesh()
        >>> op.inputs.aggressiveness.connect(my_aggressiveness)
        >>> # or
        >>> op.inputs.aggressiveness(my_aggressiveness)
        """
        return self._aggressiveness


class OutputsDecimateMesh(_Outputs):
    """Intermediate class used to get outputs from
    decimate_mesh operator.

    Examples
    --------
    >>> from ansys.dpf import core as dpf
    >>> op = dpf.operators.mesh.decimate_mesh()
    >>> # Connect inputs : op.inputs. ...
    >>> result_mesh = op.outputs.mesh()
    """

    def __init__(self, op: Operator):
        super().__init__(decimate_mesh._spec().outputs, op)
        self._mesh = Output(decimate_mesh._spec().output_pin(0), 0, op)
        self._outputs.append(self._mesh)

    @property
    def mesh(self):
        """Allows to get mesh output of the operator

        Returns
        ----------
        my_mesh : MeshedRegion

        Examples
        --------
        >>> from ansys.dpf import core as dpf
        >>> op = dpf.operators.mesh.decimate_mesh()
        >>> # Connect inputs : op.inputs. ...
        >>> result_mesh = op.outputs.mesh()
        """  # noqa: E501
        return self._mesh



"""
overall_dot
===========
Autogenerated DPF operator classes.
"""
from warnings import warn
from ansys.dpf.core.dpf_operator import Operator
from ansys.dpf.core.inputs import Input, _Inputs
from ansys.dpf.core.outputs import Output, _Outputs
from ansys.dpf.core.operators.specification import PinSpecification, Specification


class overall_dot(Operator):
    """Computes a dot product between two fields (fields are seen like a
    single large vector) and returns a scalar.

    Parameters
    ----------
    fieldA : Field
    fieldB : Field


    Examples
    --------
    >>> from ansys.dpf import core as dpf

    >>> # Instantiate operator
    >>> op = dpf.operators.math.overall_dot()

    >>> # Make input connections
    >>> my_fieldA = dpf.Field()
    >>> op.inputs.fieldA.connect(my_fieldA)
    >>> my_fieldB = dpf.Field()
    >>> op.inputs.fieldB.connect(my_fieldB)

    >>> # Instantiate operator and connect inputs in one line
    >>> op = dpf.operators.math.overall_dot(
    ...     fieldA=my_fieldA,
    ...     fieldB=my_fieldB,
    ... )

    >>> # Get output data
    >>> result_field = op.outputs.field()
    """

    def __init__(self, fieldA=None, fieldB=None, config=None, server=None):
        super().__init__(name="native::overall_dot", config=config, server=server)
        self._inputs = InputsOverallDot(self)
        self._outputs = OutputsOverallDot(self)
        if fieldA is not None:
            self.inputs.fieldA.connect(fieldA)
        if fieldB is not None:
            self.inputs.fieldB.connect(fieldB)

    @staticmethod
    def _spec():
        description = """Computes a dot product between two fields (fields are seen like a
            single large vector) and returns a scalar."""
        spec = Specification(
            description=description,
            map_input_pin_spec={
                0: PinSpecification(
                    name="fieldA",
                    type_names=["field"],
                    optional=False,
                    document="""""",
                ),
                1: PinSpecification(
                    name="fieldB",
                    type_names=["field"],
                    optional=False,
                    document="""""",
                ),
            },
            map_output_pin_spec={
                0: PinSpecification(
                    name="field",
                    type_names=["field"],
                    optional=False,
                    document="""Field defined on over-all location, contains
        a unique scalar value""",
                ),
            },
        )
        return spec

    @staticmethod
    def default_config(server=None):
        """Returns the default config of the operator.

        This config can then be changed to the user needs and be used to
        instantiate the operator. The Configuration allows to customize
        how the operation will be processed by the operator.

        Parameters
        ----------
        server : server.DPFServer, optional
            Server with channel connected to the remote or local instance. When
            ``None``, attempts to use the global server.
        """
        return Operator.default_config(name="native::overall_dot", server=server)

    @property
    def inputs(self):
        """Enables to connect inputs to the operator

        Returns
        --------
        inputs : InputsOverallDot
        """
        return super().inputs

    @property
    def outputs(self):
        """Enables to get outputs of the operator by evaluating it

        Returns
        --------
        outputs : OutputsOverallDot
        """
        return super().outputs


class InputsOverallDot(_Inputs):
    """Intermediate class used to connect user inputs to
    overall_dot operator.

    Examples
    --------
    >>> from ansys.dpf import core as dpf
    >>> op = dpf.operators.math.overall_dot()
    >>> my_fieldA = dpf.Field()
    >>> op.inputs.fieldA.connect(my_fieldA)
    >>> my_fieldB = dpf.Field()
    >>> op.inputs.fieldB.connect(my_fieldB)
    """

    def __init__(self, op: Operator):
        super().__init__(overall_dot._spec().inputs, op)
        self._fieldA = Input(overall_dot._spec().input_pin(0), 0, op, -1)
        self._inputs.append(self._fieldA)
        self._fieldB = Input(overall_dot._spec().input_pin(1), 1, op, -1)
        self._inputs.append(self._fieldB)

    @property
    def fieldA(self):
        """Allows to connect fieldA input to the operator.

        Parameters
        ----------
        my_fieldA : Field

        Examples
        --------
        >>> from ansys.dpf import core as dpf
        >>> op = dpf.operators.math.overall_dot()
        >>> op.inputs.fieldA.connect(my_fieldA)
        >>> # or
        >>> op.inputs.fieldA(my_fieldA)
        """
        return self._fieldA

    @property
    def fieldB(self):
        """Allows to connect fieldB input to the operator.

        Parameters
        ----------
        my_fieldB : Field

        Examples
        --------
        >>> from ansys.dpf import core as dpf
        >>> op = dpf.operators.math.overall_dot()
        >>> op.inputs.fieldB.connect(my_fieldB)
        >>> # or
        >>> op.inputs.fieldB(my_fieldB)
        """
        return self._fieldB


class OutputsOverallDot(_Outputs):
    """Intermediate class used to get outputs from
    overall_dot operator.

    Examples
    --------
    >>> from ansys.dpf import core as dpf
    >>> op = dpf.operators.math.overall_dot()
    >>> # Connect inputs : op.inputs. ...
    >>> result_field = op.outputs.field()
    """

    def __init__(self, op: Operator):
        super().__init__(overall_dot._spec().outputs, op)
        self._field = Output(overall_dot._spec().output_pin(0), 0, op)
        self._outputs.append(self._field)

    @property
    def field(self):
        """Allows to get field output of the operator

        Returns
        ----------
        my_field : Field

        Examples
        --------
        >>> from ansys.dpf import core as dpf
        >>> op = dpf.operators.math.overall_dot()
        >>> # Connect inputs : op.inputs. ...
        >>> result_field = op.outputs.field()
        """  # noqa: E501
        return self._field



import inspect
import sys


def _sort_supported_kwargs(bound_method, **kwargs):
    """Filters the kwargs for a given method."""
    # Ignore warnings unless specified
    if not sys.warnoptions:
        import warnings

        warnings.simplefilter("ignore")
    # Get supported arguments
    supported_args = inspect.getfullargspec(bound_method).args
    kwargs_in = {}
    kwargs_not_avail = {}
    # Filter the given arguments
    for key, item in kwargs.items():
        if key in supported_args:
            kwargs_in[key] = item
        else:
            kwargs_not_avail[key] = item
    # Prompt a warning for arguments filtered out
    if len(kwargs_not_avail) > 0:
        txt = f"The following arguments are not supported by {bound_method}: "
        txt += str(kwargs_not_avail)
        warnings.warn(txt)
    # Return the accepted arguments
    return kwargs_in



"""
elemental_mean
==============
Autogenerated DPF operator classes.
"""
from warnings import warn
from ansys.dpf.core.dpf_operator import Operator
from ansys.dpf.core.inputs import Input, _Inputs
from ansys.dpf.core.outputs import Output, _Outputs
from ansys.dpf.core.operators.specification import PinSpecification, Specification


class elemental_mean(Operator):
    """Computes the average of a multi-entity field, (ElementalNodal ->
    Elemental), (NodalElemental -> Nodal).

    Parameters
    ----------
    field : Field
    collapse_shell_layers : bool, optional
        If true, shell layers are averaged as well
        (default is false).
    force_averaging : bool, optional
        If true you average, if false you just sum.
    scoping : Scoping, optional
        Average only on these elements. if it is a
        scoping container, the label must
        correspond to the one of the fields
        containers.


    Examples
    --------
    >>> from ansys.dpf import core as dpf

    >>> # Instantiate operator
    >>> op = dpf.operators.averaging.elemental_mean()

    >>> # Make input connections
    >>> my_field = dpf.Field()
    >>> op.inputs.field.connect(my_field)
    >>> my_collapse_shell_layers = bool()
    >>> op.inputs.collapse_shell_layers.connect(my_collapse_shell_layers)
    >>> my_force_averaging = bool()
    >>> op.inputs.force_averaging.connect(my_force_averaging)
    >>> my_scoping = dpf.Scoping()
    >>> op.inputs.scoping.connect(my_scoping)

    >>> # Instantiate operator and connect inputs in one line
    >>> op = dpf.operators.averaging.elemental_mean(
    ...     field=my_field,
    ...     collapse_shell_layers=my_collapse_shell_layers,
    ...     force_averaging=my_force_averaging,
    ...     scoping=my_scoping,
    ... )

    >>> # Get output data
    >>> result_field = op.outputs.field()
    """

    def __init__(
        self,
        field=None,
        collapse_shell_layers=None,
        force_averaging=None,
        scoping=None,
        config=None,
        server=None,
    ):
        super().__init__(name="entity_average", config=config, server=server)
        self._inputs = InputsElementalMean(self)
        self._outputs = OutputsElementalMean(self)
        if field is not None:
            self.inputs.field.connect(field)
        if collapse_shell_layers is not None:
            self.inputs.collapse_shell_layers.connect(collapse_shell_layers)
        if force_averaging is not None:
            self.inputs.force_averaging.connect(force_averaging)
        if scoping is not None:
            self.inputs.scoping.connect(scoping)

    @staticmethod
    def _spec():
        description = """Computes the average of a multi-entity field, (ElementalNodal -&gt;
            Elemental), (NodalElemental -&gt; Nodal)."""
        spec = Specification(
            description=description,
            map_input_pin_spec={
                0: PinSpecification(
                    name="field",
                    type_names=["field"],
                    optional=False,
                    document="""""",
                ),
                1: PinSpecification(
                    name="collapse_shell_layers",
                    type_names=["bool"],
                    optional=True,
                    document="""If true, shell layers are averaged as well
        (default is false).""",
                ),
                2: PinSpecification(
                    name="force_averaging",
                    type_names=["bool"],
                    optional=True,
                    document="""If true you average, if false you just sum.""",
                ),
                3: PinSpecification(
                    name="scoping",
                    type_names=["scoping"],
                    optional=True,
                    document="""Average only on these elements. if it is a
        scoping container, the label must
        correspond to the one of the fields
        containers.""",
                ),
            },
            map_output_pin_spec={
                0: PinSpecification(
                    name="field",
                    type_names=["field"],
                    optional=False,
                    document="""""",
                ),
            },
        )
        return spec

    @staticmethod
    def default_config(server=None):
        """Returns the default config of the operator.

        This config can then be changed to the user needs and be used to
        instantiate the operator. The Configuration allows to customize
        how the operation will be processed by the operator.

        Parameters
        ----------
        server : server.DPFServer, optional
            Server with channel connected to the remote or local instance. When
            ``None``, attempts to use the global server.
        """
        return Operator.default_config(name="entity_average", server=server)

    @property
    def inputs(self):
        """Enables to connect inputs to the operator

        Returns
        --------
        inputs : InputsElementalMean
        """
        return super().inputs

    @property
    def outputs(self):
        """Enables to get outputs of the operator by evaluating it

        Returns
        --------
        outputs : OutputsElementalMean
        """
        return super().outputs


class InputsElementalMean(_Inputs):
    """Intermediate class used to connect user inputs to
    elemental_mean operator.

    Examples
    --------
    >>> from ansys.dpf import core as dpf
    >>> op = dpf.operators.averaging.elemental_mean()
    >>> my_field = dpf.Field()
    >>> op.inputs.field.connect(my_field)
    >>> my_collapse_shell_layers = bool()
    >>> op.inputs.collapse_shell_layers.connect(my_collapse_shell_layers)
    >>> my_force_averaging = bool()
    >>> op.inputs.force_averaging.connect(my_force_averaging)
    >>> my_scoping = dpf.Scoping()
    >>> op.inputs.scoping.connect(my_scoping)
    """

    def __init__(self, op: Operator):
        super().__init__(elemental_mean._spec().inputs, op)
        self._field = Input(elemental_mean._spec().input_pin(0), 0, op, -1)
        self._inputs.append(self._field)
        self._collapse_shell_layers = Input(
            elemental_mean._spec().input_pin(1), 1, op, -1
        )
        self._inputs.append(self._collapse_shell_layers)
        self._force_averaging = Input(elemental_mean._spec().input_pin(2), 2, op, -1)
        self._inputs.append(self._force_averaging)
        self._scoping = Input(elemental_mean._spec().input_pin(3), 3, op, -1)
        self._inputs.append(self._scoping)

    @property
    def field(self):
        """Allows to connect field input to the operator.

        Parameters
        ----------
        my_field : Field

        Examples
        --------
        >>> from ansys.dpf import core as dpf
        >>> op = dpf.operators.averaging.elemental_mean()
        >>> op.inputs.field.connect(my_field)
        >>> # or
        >>> op.inputs.field(my_field)
        """
        return self._field

    @property
    def collapse_shell_layers(self):
        """Allows to connect collapse_shell_layers input to the operator.

        If true, shell layers are averaged as well
        (default is false).

        Parameters
        ----------
        my_collapse_shell_layers : bool

        Examples
        --------
        >>> from ansys.dpf import core as dpf
        >>> op = dpf.operators.averaging.elemental_mean()
        >>> op.inputs.collapse_shell_layers.connect(my_collapse_shell_layers)
        >>> # or
        >>> op.inputs.collapse_shell_layers(my_collapse_shell_layers)
        """
        return self._collapse_shell_layers

    @property
    def force_averaging(self):
        """Allows to connect force_averaging input to the operator.

        If true you average, if false you just sum.

        Parameters
        ----------
        my_force_averaging : bool

        Examples
        --------
        >>> from ansys.dpf import core as dpf
        >>> op = dpf.operators.averaging.elemental_mean()
        >>> op.inputs.force_averaging.connect(my_force_averaging)
        >>> # or
        >>> op.inputs.force_averaging(my_force_averaging)
        """
        return self._force_averaging

    @property
    def scoping(self):
        """Allows to connect scoping input to the operator.

        Average only on these elements. if it is a
        scoping container, the label must
        correspond to the one of the fields
        containers.

        Parameters
        ----------
        my_scoping : Scoping

        Examples
        --------
        >>> from ansys.dpf import core as dpf
        >>> op = dpf.operators.averaging.elemental_mean()
        >>> op.inputs.scoping.connect(my_scoping)
        >>> # or
        >>> op.inputs.scoping(my_scoping)
        """
        return self._scoping


class OutputsElementalMean(_Outputs):
    """Intermediate class used to get outputs from
    elemental_mean operator.

    Examples
    --------
    >>> from ansys.dpf import core as dpf
    >>> op = dpf.operators.averaging.elemental_mean()
    >>> # Connect inputs : op.inputs. ...
    >>> result_field = op.outputs.field()
    """

    def __init__(self, op: Operator):
        super().__init__(elemental_mean._spec().outputs, op)
        self._field = Output(elemental_mean._spec().output_pin(0), 0, op)
        self._outputs.append(self._field)

    @property
    def field(self):
        """Allows to get field output of the operator

        Returns
        ----------
        my_field : Field

        Examples
        --------
        >>> from ansys.dpf import core as dpf
        >>> op = dpf.operators.averaging.elemental_mean()
        >>> # Connect inputs : op.inputs. ...
        >>> result_field = op.outputs.field()
        """  # noqa: E501
        return self._field



"""
to_polar_coordinates
====================
Autogenerated DPF operator classes.
"""
from warnings import warn
from ansys.dpf.core.dpf_operator import Operator
from ansys.dpf.core.inputs import Input, _Inputs
from ansys.dpf.core.outputs import Output, _Outputs
from ansys.dpf.core.operators.specification import PinSpecification, Specification


class to_polar_coordinates(Operator):
    """Finds r, theta (rad), and z coordinates of a coordinates (nodal) field
    in a cartesian coordinates system where the input coordinate
    system defines the rotation axis and the origin.

    Parameters
    ----------
    field : Field or FieldsContainer
        Field or fields container with only one field
        is expected
    coordinate_system : Field, optional
        3-3 rotation matrix and origin coordinates
        must be set here to define a
        coordinate system. by default, the
        rotation axis is the z axis and the
        origin is [0,0,0].


    Examples
    --------
    >>> from ansys.dpf import core as dpf

    >>> # Instantiate operator
    >>> op = dpf.operators.geo.to_polar_coordinates()

    >>> # Make input connections
    >>> my_field = dpf.Field()
    >>> op.inputs.field.connect(my_field)
    >>> my_coordinate_system = dpf.Field()
    >>> op.inputs.coordinate_system.connect(my_coordinate_system)

    >>> # Instantiate operator and connect inputs in one line
    >>> op = dpf.operators.geo.to_polar_coordinates(
    ...     field=my_field,
    ...     coordinate_system=my_coordinate_system,
    ... )

    >>> # Get output data
    >>> result_field = op.outputs.field()
    """

    def __init__(self, field=None, coordinate_system=None, config=None, server=None):
        super().__init__(name="polar_coordinates", config=config, server=server)
        self._inputs = InputsToPolarCoordinates(self)
        self._outputs = OutputsToPolarCoordinates(self)
        if field is not None:
            self.inputs.field.connect(field)
        if coordinate_system is not None:
            self.inputs.coordinate_system.connect(coordinate_system)

    @staticmethod
    def _spec():
        description = """Finds r, theta (rad), and z coordinates of a coordinates (nodal) field
            in a cartesian coordinates system where the input
            coordinate system defines the rotation axis and the
            origin."""
        spec = Specification(
            description=description,
            map_input_pin_spec={
                0: PinSpecification(
                    name="field",
                    type_names=["field", "fields_container"],
                    optional=False,
                    document="""Field or fields container with only one field
        is expected""",
                ),
                1: PinSpecification(
                    name="coordinate_system",
                    type_names=["field"],
                    optional=True,
                    document="""3-3 rotation matrix and origin coordinates
        must be set here to define a
        coordinate system. by default, the
        rotation axis is the z axis and the
        origin is [0,0,0].""",
                ),
            },
            map_output_pin_spec={
                0: PinSpecification(
                    name="field",
                    type_names=["field"],
                    optional=False,
                    document="""""",
                ),
            },
        )
        return spec

    @staticmethod
    def default_config(server=None):
        """Returns the default config of the operator.

        This config can then be changed to the user needs and be used to
        instantiate the operator. The Configuration allows to customize
        how the operation will be processed by the operator.

        Parameters
        ----------
        server : server.DPFServer, optional
            Server with channel connected to the remote or local instance. When
            ``None``, attempts to use the global server.
        """
        return Operator.default_config(name="polar_coordinates", server=server)

    @property
    def inputs(self):
        """Enables to connect inputs to the operator

        Returns
        --------
        inputs : InputsToPolarCoordinates
        """
        return super().inputs

    @property
    def outputs(self):
        """Enables to get outputs of the operator by evaluating it

        Returns
        --------
        outputs : OutputsToPolarCoordinates
        """
        return super().outputs


class InputsToPolarCoordinates(_Inputs):
    """Intermediate class used to connect user inputs to
    to_polar_coordinates operator.

    Examples
    --------
    >>> from ansys.dpf import core as dpf
    >>> op = dpf.operators.geo.to_polar_coordinates()
    >>> my_field = dpf.Field()
    >>> op.inputs.field.connect(my_field)
    >>> my_coordinate_system = dpf.Field()
    >>> op.inputs.coordinate_system.connect(my_coordinate_system)
    """

    def __init__(self, op: Operator):
        super().__init__(to_polar_coordinates._spec().inputs, op)
        self._field = Input(to_polar_coordinates._spec().input_pin(0), 0, op, -1)
        self._inputs.append(self._field)
        self._coordinate_system = Input(
            to_polar_coordinates._spec().input_pin(1), 1, op, -1
        )
        self._inputs.append(self._coordinate_system)

    @property
    def field(self):
        """Allows to connect field input to the operator.

        Field or fields container with only one field
        is expected

        Parameters
        ----------
        my_field : Field or FieldsContainer

        Examples
        --------
        >>> from ansys.dpf import core as dpf
        >>> op = dpf.operators.geo.to_polar_coordinates()
        >>> op.inputs.field.connect(my_field)
        >>> # or
        >>> op.inputs.field(my_field)
        """
        return self._field

    @property
    def coordinate_system(self):
        """Allows to connect coordinate_system input to the operator.

        3-3 rotation matrix and origin coordinates
        must be set here to define a
        coordinate system. by default, the
        rotation axis is the z axis and the
        origin is [0,0,0].

        Parameters
        ----------
        my_coordinate_system : Field

        Examples
        --------
        >>> from ansys.dpf import core as dpf
        >>> op = dpf.operators.geo.to_polar_coordinates()
        >>> op.inputs.coordinate_system.connect(my_coordinate_system)
        >>> # or
        >>> op.inputs.coordinate_system(my_coordinate_system)
        """
        return self._coordinate_system


class OutputsToPolarCoordinates(_Outputs):
    """Intermediate class used to get outputs from
    to_polar_coordinates operator.

    Examples
    --------
    >>> from ansys.dpf import core as dpf
    >>> op = dpf.operators.geo.to_polar_coordinates()
    >>> # Connect inputs : op.inputs. ...
    >>> result_field = op.outputs.field()
    """

    def __init__(self, op: Operator):
        super().__init__(to_polar_coordinates._spec().outputs, op)
        self._field = Output(to_polar_coordinates._spec().output_pin(0), 0, op)
        self._outputs.append(self._field)

    @property
    def field(self):
        """Allows to get field output of the operator

        Returns
        ----------
        my_field : Field

        Examples
        --------
        >>> from ansys.dpf import core as dpf
        >>> op = dpf.operators.geo.to_polar_coordinates()
        >>> # Connect inputs : op.inputs. ...
        >>> result_field = op.outputs.field()
        """  # noqa: E501
        return self._field



"""
transpose
=========
Autogenerated DPF operator classes.
"""
from warnings import warn
from ansys.dpf.core.dpf_operator import Operator
from ansys.dpf.core.inputs import Input, _Inputs
from ansys.dpf.core.outputs import Output, _Outputs
from ansys.dpf.core.outputs import _modify_output_spec_with_one_type
from ansys.dpf.core.operators.specification import PinSpecification, Specification


class transpose(Operator):
    """Transposes the input scoping or scopings container (Elemental/Faces
    --> Nodal, or Nodal ---> Elemental/Faces), based on the input mesh
    region.

    Parameters
    ----------
    mesh_scoping : Scoping or ScopingsContainer
        Scoping or scopings container (the input type
        is the output type)
    meshed_region : MeshedRegion or MeshesContainer
    inclusive : int, optional
        If inclusive == 1 then all the elements/faces
        adjacent to the nodes ids in input
        are added, if inclusive == 0, only
        the elements/faces which have all
        their nodes in the scoping are
        included
    requested_location : str, optional
        Output scoping location for meshes with
        nodes, faces and elements. by
        default, elemental and faces scopings
        transpose to nodal, and nodal
        scopings transpose to elemental.


    Examples
    --------
    >>> from ansys.dpf import core as dpf

    >>> # Instantiate operator
    >>> op = dpf.operators.scoping.transpose()

    >>> # Make input connections
    >>> my_mesh_scoping = dpf.Scoping()
    >>> op.inputs.mesh_scoping.connect(my_mesh_scoping)
    >>> my_meshed_region = dpf.MeshedRegion()
    >>> op.inputs.meshed_region.connect(my_meshed_region)
    >>> my_inclusive = int()
    >>> op.inputs.inclusive.connect(my_inclusive)
    >>> my_requested_location = str()
    >>> op.inputs.requested_location.connect(my_requested_location)

    >>> # Instantiate operator and connect inputs in one line
    >>> op = dpf.operators.scoping.transpose(
    ...     mesh_scoping=my_mesh_scoping,
    ...     meshed_region=my_meshed_region,
    ...     inclusive=my_inclusive,
    ...     requested_location=my_requested_location,
    ... )

    >>> # Get output data
    >>> result_mesh_scoping = op.outputs.mesh_scoping()
    """

    def __init__(
        self,
        mesh_scoping=None,
        meshed_region=None,
        inclusive=None,
        requested_location=None,
        config=None,
        server=None,
    ):
        super().__init__(name="transpose_scoping", config=config, server=server)
        self._inputs = InputsTranspose(self)
        self._outputs = OutputsTranspose(self)
        if mesh_scoping is not None:
            self.inputs.mesh_scoping.connect(mesh_scoping)
        if meshed_region is not None:
            self.inputs.meshed_region.connect(meshed_region)
        if inclusive is not None:
            self.inputs.inclusive.connect(inclusive)
        if requested_location is not None:
            self.inputs.requested_location.connect(requested_location)

    @staticmethod
    def _spec():
        description = """Transposes the input scoping or scopings container (Elemental/Faces
            --&gt; Nodal, or Nodal ---&gt; Elemental/Faces), based on the
            input mesh region."""
        spec = Specification(
            description=description,
            map_input_pin_spec={
                0: PinSpecification(
                    name="mesh_scoping",
                    type_names=["scoping", "scopings_container"],
                    optional=False,
                    document="""Scoping or scopings container (the input type
        is the output type)""",
                ),
                1: PinSpecification(
                    name="meshed_region",
                    type_names=["meshed_region", "meshes_container"],
                    optional=False,
                    document="""""",
                ),
                2: PinSpecification(
                    name="inclusive",
                    type_names=["int32"],
                    optional=True,
                    document="""If inclusive == 1 then all the elements/faces
        adjacent to the nodes ids in input
        are added, if inclusive == 0, only
        the elements/faces which have all
        their nodes in the scoping are
        included""",
                ),
                9: PinSpecification(
                    name="requested_location",
                    type_names=["string"],
                    optional=True,
                    document="""Output scoping location for meshes with
        nodes, faces and elements. by
        default, elemental and faces scopings
        transpose to nodal, and nodal
        scopings transpose to elemental.""",
                ),
            },
            map_output_pin_spec={
                0: PinSpecification(
                    name="mesh_scoping",
                    type_names=["scoping", "scopings_container"],
                    optional=False,
                    document="""Scoping or scopings container (the input type
        is the output type)""",
                ),
            },
        )
        return spec

    @staticmethod
    def default_config(server=None):
        """Returns the default config of the operator.

        This config can then be changed to the user needs and be used to
        instantiate the operator. The Configuration allows to customize
        how the operation will be processed by the operator.

        Parameters
        ----------
        server : server.DPFServer, optional
            Server with channel connected to the remote or local instance. When
            ``None``, attempts to use the global server.
        """
        return Operator.default_config(name="transpose_scoping", server=server)

    @property
    def inputs(self):
        """Enables to connect inputs to the operator

        Returns
        --------
        inputs : InputsTranspose
        """
        return super().inputs

    @property
    def outputs(self):
        """Enables to get outputs of the operator by evaluating it

        Returns
        --------
        outputs : OutputsTranspose
        """
        return super().outputs


class InputsTranspose(_Inputs):
    """Intermediate class used to connect user inputs to
    transpose operator.

    Examples
    --------
    >>> from ansys.dpf import core as dpf
    >>> op = dpf.operators.scoping.transpose()
    >>> my_mesh_scoping = dpf.Scoping()
    >>> op.inputs.mesh_scoping.connect(my_mesh_scoping)
    >>> my_meshed_region = dpf.MeshedRegion()
    >>> op.inputs.meshed_region.connect(my_meshed_region)
    >>> my_inclusive = int()
    >>> op.inputs.inclusive.connect(my_inclusive)
    >>> my_requested_location = str()
    >>> op.inputs.requested_location.connect(my_requested_location)
    """

    def __init__(self, op: Operator):
        super().__init__(transpose._spec().inputs, op)
        self._mesh_scoping = Input(transpose._spec().input_pin(0), 0, op, -1)
        self._inputs.append(self._mesh_scoping)
        self._meshed_region = Input(transpose._spec().input_pin(1), 1, op, -1)
        self._inputs.append(self._meshed_region)
        self._inclusive = Input(transpose._spec().input_pin(2), 2, op, -1)
        self._inputs.append(self._inclusive)
        self._requested_location = Input(transpose._spec().input_pin(9), 9, op, -1)
        self._inputs.append(self._requested_location)

    @property
    def mesh_scoping(self):
        """Allows to connect mesh_scoping input to the operator.

        Scoping or scopings container (the input type
        is the output type)

        Parameters
        ----------
        my_mesh_scoping : Scoping or ScopingsContainer

        Examples
        --------
        >>> from ansys.dpf import core as dpf
        >>> op = dpf.operators.scoping.transpose()
        >>> op.inputs.mesh_scoping.connect(my_mesh_scoping)
        >>> # or
        >>> op.inputs.mesh_scoping(my_mesh_scoping)
        """
        return self._mesh_scoping

    @property
    def meshed_region(self):
        """Allows to connect meshed_region input to the operator.

        Parameters
        ----------
        my_meshed_region : MeshedRegion or MeshesContainer

        Examples
        --------
        >>> from ansys.dpf import core as dpf
        >>> op = dpf.operators.scoping.transpose()
        >>> op.inputs.meshed_region.connect(my_meshed_region)
        >>> # or
        >>> op.inputs.meshed_region(my_meshed_region)
        """
        return self._meshed_region

    @property
    def inclusive(self):
        """Allows to connect inclusive input to the operator.

        If inclusive == 1 then all the elements/faces
        adjacent to the nodes ids in input
        are added, if inclusive == 0, only
        the elements/faces which have all
        their nodes in the scoping are
        included

        Parameters
        ----------
        my_inclusive : int

        Examples
        --------
        >>> from ansys.dpf import core as dpf
        >>> op = dpf.operators.scoping.transpose()
        >>> op.inputs.inclusive.connect(my_inclusive)
        >>> # or
        >>> op.inputs.inclusive(my_inclusive)
        """
        return self._inclusive

    @property
    def requested_location(self):
        """Allows to connect requested_location input to the operator.

        Output scoping location for meshes with
        nodes, faces and elements. by
        default, elemental and faces scopings
        transpose to nodal, and nodal
        scopings transpose to elemental.

        Parameters
        ----------
        my_requested_location : str

        Examples
        --------
        >>> from ansys.dpf import core as dpf
        >>> op = dpf.operators.scoping.transpose()
        >>> op.inputs.requested_location.connect(my_requested_location)
        >>> # or
        >>> op.inputs.requested_location(my_requested_location)
        """
        return self._requested_location


class OutputsTranspose(_Outputs):
    """Intermediate class used to get outputs from
    transpose operator.

    Examples
    --------
    >>> from ansys.dpf import core as dpf
    >>> op = dpf.operators.scoping.transpose()
    >>> # Connect inputs : op.inputs. ...
    >>> result_mesh_scoping = op.outputs.mesh_scoping()
    """

    def __init__(self, op: Operator):
        super().__init__(transpose._spec().outputs, op)
        self.mesh_scoping_as_scoping = Output(
            _modify_output_spec_with_one_type(
                transpose._spec().output_pin(0), "scoping"
            ),
            0,
            op,
        )
        self._outputs.append(self.mesh_scoping_as_scoping)
        self.mesh_scoping_as_scopings_container = Output(
            _modify_output_spec_with_one_type(
                transpose._spec().output_pin(0), "scopings_container"
            ),
            0,
            op,
        )
        self._outputs.append(self.mesh_scoping_as_scopings_container)



"""
merge_scopings
==============
Autogenerated DPF operator classes.
"""
from warnings import warn
from ansys.dpf.core.dpf_operator import Operator
from ansys.dpf.core.inputs import Input, _Inputs
from ansys.dpf.core.outputs import Output, _Outputs
from ansys.dpf.core.operators.specification import PinSpecification, Specification


class merge_scopings(Operator):
    """Assembles a set of scopings into a unique one.

    Parameters
    ----------
    scopings1 : Scoping or ScopingsContainer
        Either a scopings container, a vector of
        scopings to merge, or scopings from
        pin 0 to ...
    scopings2 : Scoping or ScopingsContainer
        Either a scopings container, a vector of
        scopings to merge, or scopings from
        pin 0 to ...


    Examples
    --------
    >>> from ansys.dpf import core as dpf

    >>> # Instantiate operator
    >>> op = dpf.operators.utility.merge_scopings()

    >>> # Make input connections
    >>> my_scopings1 = dpf.Scoping()
    >>> op.inputs.scopings1.connect(my_scopings1)
    >>> my_scopings2 = dpf.Scoping()
    >>> op.inputs.scopings2.connect(my_scopings2)

    >>> # Instantiate operator and connect inputs in one line
    >>> op = dpf.operators.utility.merge_scopings(
    ...     scopings1=my_scopings1,
    ...     scopings2=my_scopings2,
    ... )

    >>> # Get output data
    >>> result_merged_scoping = op.outputs.merged_scoping()
    """

    def __init__(self, scopings1=None, scopings2=None, config=None, server=None):
        super().__init__(name="merge::scoping", config=config, server=server)
        self._inputs = InputsMergeScopings(self)
        self._outputs = OutputsMergeScopings(self)
        if scopings1 is not None:
            self.inputs.scopings1.connect(scopings1)
        if scopings2 is not None:
            self.inputs.scopings2.connect(scopings2)

    @staticmethod
    def _spec():
        description = """Assembles a set of scopings into a unique one."""
        spec = Specification(
            description=description,
            map_input_pin_spec={
                0: PinSpecification(
                    name="scopings",
                    type_names=["scoping", "scopings_container"],
                    optional=False,
                    document="""Either a scopings container, a vector of
        scopings to merge, or scopings from
        pin 0 to ...""",
                ),
                1: PinSpecification(
                    name="scopings",
                    type_names=["scoping", "scopings_container"],
                    optional=False,
                    document="""Either a scopings container, a vector of
        scopings to merge, or scopings from
        pin 0 to ...""",
                ),
            },
            map_output_pin_spec={
                0: PinSpecification(
                    name="merged_scoping",
                    type_names=["scoping"],
                    optional=False,
                    document="""""",
                ),
            },
        )
        return spec

    @staticmethod
    def default_config(server=None):
        """Returns the default config of the operator.

        This config can then be changed to the user needs and be used to
        instantiate the operator. The Configuration allows to customize
        how the operation will be processed by the operator.

        Parameters
        ----------
        server : server.DPFServer, optional
            Server with channel connected to the remote or local instance. When
            ``None``, attempts to use the global server.
        """
        return Operator.default_config(name="merge::scoping", server=server)

    @property
    def inputs(self):
        """Enables to connect inputs to the operator

        Returns
        --------
        inputs : InputsMergeScopings
        """
        return super().inputs

    @property
    def outputs(self):
        """Enables to get outputs of the operator by evaluating it

        Returns
        --------
        outputs : OutputsMergeScopings
        """
        return super().outputs


class InputsMergeScopings(_Inputs):
    """Intermediate class used to connect user inputs to
    merge_scopings operator.

    Examples
    --------
    >>> from ansys.dpf import core as dpf
    >>> op = dpf.operators.utility.merge_scopings()
    >>> my_scopings1 = dpf.Scoping()
    >>> op.inputs.scopings1.connect(my_scopings1)
    >>> my_scopings2 = dpf.Scoping()
    >>> op.inputs.scopings2.connect(my_scopings2)
    """

    def __init__(self, op: Operator):
        super().__init__(merge_scopings._spec().inputs, op)
        self._scopings1 = Input(merge_scopings._spec().input_pin(0), 0, op, 0)
        self._inputs.append(self._scopings1)
        self._scopings2 = Input(merge_scopings._spec().input_pin(1), 1, op, 1)
        self._inputs.append(self._scopings2)

    @property
    def scopings1(self):
        """Allows to connect scopings1 input to the operator.

        Either a scopings container, a vector of
        scopings to merge, or scopings from
        pin 0 to ...

        Parameters
        ----------
        my_scopings1 : Scoping or ScopingsContainer

        Examples
        --------
        >>> from ansys.dpf import core as dpf
        >>> op = dpf.operators.utility.merge_scopings()
        >>> op.inputs.scopings1.connect(my_scopings1)
        >>> # or
        >>> op.inputs.scopings1(my_scopings1)
        """
        return self._scopings1

    @property
    def scopings2(self):
        """Allows to connect scopings2 input to the operator.

        Either a scopings container, a vector of
        scopings to merge, or scopings from
        pin 0 to ...

        Parameters
        ----------
        my_scopings2 : Scoping or ScopingsContainer

        Examples
        --------
        >>> from ansys.dpf import core as dpf
        >>> op = dpf.operators.utility.merge_scopings()
        >>> op.inputs.scopings2.connect(my_scopings2)
        >>> # or
        >>> op.inputs.scopings2(my_scopings2)
        """
        return self._scopings2


class OutputsMergeScopings(_Outputs):
    """Intermediate class used to get outputs from
    merge_scopings operator.

    Examples
    --------
    >>> from ansys.dpf import core as dpf
    >>> op = dpf.operators.utility.merge_scopings()
    >>> # Connect inputs : op.inputs. ...
    >>> result_merged_scoping = op.outputs.merged_scoping()
    """

    def __init__(self, op: Operator):
        super().__init__(merge_scopings._spec().outputs, op)
        self._merged_scoping = Output(merge_scopings._spec().output_pin(0), 0, op)
        self._outputs.append(self._merged_scoping)

    @property
    def merged_scoping(self):
        """Allows to get merged_scoping output of the operator

        Returns
        ----------
        my_merged_scoping : Scoping

        Examples
        --------
        >>> from ansys.dpf import core as dpf
        >>> op = dpf.operators.utility.merge_scopings()
        >>> # Connect inputs : op.inputs. ...
        >>> result_merged_scoping = op.outputs.merged_scoping()
        """  # noqa: E501
        return self._merged_scoping



"""
Server types
============
Contains the different kinds of
servers available for the factory.
"""
import abc
import io
import os
import socket
import subprocess
import time
import warnings
import traceback
from threading import Thread, Lock
from abc import ABC

import psutil

import ansys.dpf.core as core
from ansys.dpf.core.check_version import server_meet_version
from ansys.dpf.core import errors, server_factory
from ansys.dpf.core._version import (
    server_to_ansys_grpc_dpf_version,
    server_to_ansys_version,
)
from ansys.dpf.core import server_context
from ansys.dpf.gate import load_api, data_processing_grpcapi

import logging

LOG = logging.getLogger(__name__)
LOG.setLevel("DEBUG")
DPF_DEFAULT_PORT = int(os.environ.get("DPF_PORT", 50054))
LOCALHOST = os.environ.get("DPF_IP", "127.0.0.1")
RUNNING_DOCKER = server_factory.create_default_docker_config()

MAX_PORT = 65535


def _get_dll_path(name, ansys_path=None):
    """Helper function to get the right dll path for Linux or Windows"""
    ISPOSIX = os.name == "posix"
    ANSYS_INSTALL = core.misc.get_ansys_path(ansys_path)
    api_path = load_api._get_path_in_install()
    if api_path is None:
        raise ImportError(f"Could not find API path in install.")
    SUB_FOLDERS = os.path.join(ANSYS_INSTALL, api_path)
    if ISPOSIX:
        name = "lib" + name
    return os.path.join(SUB_FOLDERS, name)


def check_valid_ip(ip):
    """Check if a valid IP address is entered.

    This method raises an error when an invalid IP address is entered.
    """
    try:
        socket.inet_aton(ip)
    except OSError:
        raise ValueError(f'Invalid IP address "{ip}"')


def _verify_ansys_path_is_valid(ansys_path, executable, path_in_install=None):
    if path_in_install is None:
        path_in_install = load_api._get_path_in_install()
    if os.path.isdir(f"{ansys_path}/{path_in_install}"):
        dpf_run_dir = f"{ansys_path}/{path_in_install}"
    else:
        dpf_run_dir = f"{ansys_path}"
    if not os.path.isdir(dpf_run_dir):
        raise NotADirectoryError(
            f'Invalid ansys path at "{ansys_path}".  '
            "Unable to locate the directory containing DPF at "
            f'"{dpf_run_dir}"'
        )
    else:
        if not os.path.exists(os.path.join(dpf_run_dir, executable)):
            raise FileNotFoundError(
                f'DPF executable not found at "{dpf_run_dir}".  '
                f'Unable to locate the executable "{executable}"'
            )
    return dpf_run_dir


def _run_launch_server_process(
    ip, port, ansys_path=None, docker_config=server_factory.RunningDockerConfig()
):
    bShell = False
    if docker_config.use_docker:
        docker_server_port = int(os.environ.get("DOCKER_SERVER_PORT", port))
        dpf_run_dir = os.getcwd()
        if os.name == "posix":
            bShell = True
        run_cmd = docker_config.docker_run_cmd_command(docker_server_port, port)
    else:
        if os.name == "nt":
            executable = "Ans.Dpf.Grpc.bat"
            run_cmd = f"{executable} --address {ip} --port {port}"
        else:
            executable = "./Ans.Dpf.Grpc.sh"  # pragma: no cover
            run_cmd = [
                executable,
                f"--address {ip}",
                f"--port {port}",
            ]  # pragma: no cover
        path_in_install = load_api._get_path_in_install(internal_folder="bin")
        dpf_run_dir = _verify_ansys_path_is_valid(ansys_path, executable, path_in_install)

    old_dir = os.getcwd()
    os.chdir(dpf_run_dir)
    if not bShell:
        process = subprocess.Popen(run_cmd, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
    else:
        process = subprocess.Popen(
            run_cmd, stdout=subprocess.PIPE, stderr=subprocess.PIPE, shell=True
        )
    os.chdir(old_dir)
    return process


def _wait_and_check_server_connection(
    process, port, timeout, lines, current_errors, stderr=None, stdout=None
):
    if not stderr:

        def read_stderr():
            with io.TextIOWrapper(process.stderr, encoding="utf-8") as log_err:
                for line in log_err:
                    LOG.error(line)
                    current_errors.append(line)

        stderr = read_stderr
        # check to see if the service started
    if not stdout:

        def read_stdout():
            with io.TextIOWrapper(process.stdout, encoding="utf-8") as log_out:
                for line in log_out:
                    LOG.debug(line)
                    lines.append(line)

        stdout = read_stdout

    # must be in the background since the process reader is blocking
    Thread(target=stdout, daemon=True).start()
    Thread(target=stderr, daemon=True).start()

    t_timeout = time.time() + timeout
    started = False
    timedout = False
    while not started and len(current_errors) == 0:
        # print(lines)
        started = any("server started" in line for line in lines)

        if time.time() > t_timeout:
            if timedout:
                raise TimeoutError(f"Server did not start in {timeout + timeout} seconds")
            timedout = True
            t_timeout += timeout

    # verify there were no errors
    time.sleep(0.01)
    if current_errors:
        try:
            process.kill()
        except PermissionError:
            pass
        errstr = "\n".join(current_errors)
        if (
            "Only one usage of each socket address" in errstr
            or "port is already allocated" in errstr
        ):
            raise errors.InvalidPortError(f"Port {port} in use")
        raise RuntimeError(errstr)


def launch_dpf(ansys_path, ip=LOCALHOST, port=DPF_DEFAULT_PORT, timeout=10):
    """Launch Ansys DPF.

    Parameters
    ----------
    ansys_path : str, optional
        Root path for the Ansys installation directory. For example, ``"/ansys_inc/v212/"``.
        The default is the latest Ansys installation.
    ip : str, optional
        IP address of the remote or local instance to connect to. The
        default is ``"LOCALHOST"``.
    port : int
        Port to connect to the remote instance on. The default is
        ``"DPF_DEFAULT_PORT"``, which is 50054.
    timeout : float, optional
        Maximum number of seconds for the initialization attempt.
        The default is ``10``. Once the specified number of seconds
        passes, the connection fails.

    """
    process = _run_launch_server_process(ip, port, ansys_path)
    lines = []
    current_errors = []
    _wait_and_check_server_connection(
        process, port, timeout, lines, current_errors, stderr=None, stdout=None
    )


def launch_dpf_on_docker(
    running_docker_config=server_factory.RunningDockerConfig(),
    ansys_path=None,
    ip=LOCALHOST,
    port=DPF_DEFAULT_PORT,
    timeout=10.0,
):
    """Launch Ansys DPF.

    Parameters
    ----------
    running_docker_config : server_factory.RunningDockerConfig, optional
        To start DPF server as a docker, specify the docker configurations here.
    ansys_path : str, optional
        Root path for the Ansys installation directory. For example, ``"/ansys_inc/v212/"``.
        The default is the latest Ansys installation.
    ip : str, optional
        IP address of the remote or local instance to connect to. The
        default is ``"LOCALHOST"``.
    port : int
        Port to connect to the remote instance on. The default is
        ``"DPF_DEFAULT_PORT"``, which is 50054.
    timeout : float, optional
        Maximum number of seconds for the initialization attempt.
        The default is ``10``. Once the specified number of seconds
        passes, the connection fails.

    """
    process = _run_launch_server_process(ip, port, ansys_path, running_docker_config)

    # check to see if the service started
    cmd_lines = []
    # Creating lock for threads
    lock = Lock()
    lock.acquire()
    lines = []
    current_errors = []
    running_docker_config.docker_server_port = port

    def read_stdout():
        with io.TextIOWrapper(process.stdout, encoding="utf-8") as log_out:
            for line in log_out:
                LOG.debug(line)
                cmd_lines.append(line)
                lock.release()
            running_docker_config.listen_to_process(LOG, cmd_lines, lines, timeout)

    def read_stderr():
        with io.TextIOWrapper(process.stderr, encoding="utf-8") as log_err:
            for line in log_err:
                LOG.error(line)
                current_errors.append(line)
            while lock.locked():
                pass
            running_docker_config.listen_to_process(LOG, cmd_lines, current_errors, timeout, False)

    _wait_and_check_server_connection(
        process,
        port,
        timeout,
        lines,
        current_errors,
        stderr=read_stderr,
        stdout=read_stdout,
    )


def launch_remote_dpf(version=None):
    try:
        import ansys.platform.instancemanagement as pypim
    except ImportError as e:
        raise ImportError(
            "Launching a remote session of DPF requires the installation"
            + " of ansys-platform-instancemanagement"
        ) from e
    pim = pypim.connect()

    # Possible improvement:
    # When the version is not specified, it would be possible to use
    # pim.list_definition(product_name="dpf") and select a version compatible with the current
    # pydpf version, following the compatibility rules
    instance = pim.create_instance(product_name="dpf", product_version=version)
    instance.wait_for_ready()
    grpc_service = instance.services["grpc"]
    if grpc_service.headers:
        LOG.error(
            "Communicating with DPF in this remote environment requires metadata."
            + "This is not supported, you will likely encounter errors or limitations."
        )
    return instance


def _compare_ansys_grpc_dpf_version(right_grpc_module_version_str: str, grpc_module_version: str):
    if right_grpc_module_version_str:
        import re
        from packaging.version import parse as parse_version

        right_version_first_numbers = re.search(r"\d", right_grpc_module_version_str)
        right_version_numbers = right_grpc_module_version_str[right_version_first_numbers.start() :]
        compare = (
            "=="
            if right_version_first_numbers.start() == 0
            else right_grpc_module_version_str[0 : right_version_first_numbers.start()].strip()
        )
        if compare == "==":
            return parse_version(grpc_module_version) == parse_version(right_version_numbers)
        elif compare == ">=":
            return parse_version(grpc_module_version) >= parse_version(right_version_numbers)
        elif compare == ">":
            return parse_version(grpc_module_version) > parse_version(right_version_numbers)
        elif compare == "<=":
            return parse_version(grpc_module_version) <= parse_version(right_version_numbers)
        elif compare == "<":
            return parse_version(grpc_module_version) < parse_version(right_version_numbers)
    return True


def check_ansys_grpc_dpf_version(server, timeout):
    import ansys.grpc.dpf
    import grpc

    state = grpc.channel_ready_future(server.channel)
    # verify connection has matured
    tstart = time.time()
    while ((time.time() - tstart) < timeout) and not state._matured:
        time.sleep(0.001)

    if not state._matured:
        raise TimeoutError(
            f"Failed to connect to {server._input_ip}:{server._input_port} in {timeout} seconds"
        )
    compatibility_link = (
        f"https://dpf.docs.pyansys.com/getting_started/" f"index.html#client-server-compatibility"
    )
    LOG.debug("Established connection to DPF gRPC")
    grpc_module_version = ansys.grpc.dpf.__version__
    server_version = server.version
    right_grpc_module_version = server_to_ansys_grpc_dpf_version.get(server_version, None)
    if right_grpc_module_version is None:  # pragma: no cover
        # warnings.warn(f"No requirement specified on ansys-grpc-dpf for server version "
        #               f"{server_version}. Continuing with the ansys-grpc-dpf version "
        #               f"installed ({grpc_module_version}). In case of unexpected instability, "
        #               f"please refer to the compatibility guidelines given in "
        #               f"{compatibility_link}.")
        return
    if not _compare_ansys_grpc_dpf_version(right_grpc_module_version, grpc_module_version):
        ansys_version_to_use = server_to_ansys_version.get(server_version, "Unknown")
        ansys_versions = core._version.server_to_ansys_version
        latest_ansys = ansys_versions[max(ansys_versions.keys())]
        raise ImportWarning(
            f"An incompatibility has been detected between the DPF server version "
            f"({server_version} "
            f"from Ansys {ansys_version_to_use})"
            f" and the ansys-grpc-dpf version installed ({grpc_module_version})."
            f" Please consider using the latest DPF server available in the "
            f"{latest_ansys} Ansys unified install.\n"
            f"To follow the compatibility guidelines given in "
            f"{compatibility_link} while still using DPF server {server_version}, "
            f"please install version {right_grpc_module_version} of ansys-grpc-dpf"
            f" with the command: \n"
            f"     pip install ansys-grpc-dpf{right_grpc_module_version}"
        )


class BaseServer(abc.ABC):
    """Abstract class for servers"""

    @abc.abstractmethod
    def __init__(self):
        """Base class for all types of servers: grpc, in process..."""
        # TODO: Use _server_id to compare servers for equality?
        self._server_id = None
        self._session_instance = None
        self._base_service_instance = None
        self._context = None
        self._docker_config = server_factory.RunningDockerConfig()

    def set_as_global(self, as_global=True):
        """Set the current server as global if necessary.

        Parameters
        ----------
        as_global : bool, optional
            Global variable that stores the IP address and port for the DPF
            module. All DPF objects created in this Python session will
            use this IP and port. The default is ``True``.
        """
        # assign to global channel when requested
        if as_global:
            core.SERVER = self

    def has_client(self):
        return not (self.client is None)

    @property
    @abc.abstractmethod
    def client(self):
        pass

    @property
    @abc.abstractmethod
    def version(self):
        pass

    @property
    @abc.abstractmethod
    def available_api_types(self):
        pass

    @abc.abstractmethod
    def get_api_for_type(self, c_api, grpc_api):
        pass

    @property
    def info(self):
        """Server information.

        Returns
        -------
        info : dictionary
            Dictionary with server information, including ``"server_ip"``,
            ``"server_port"``, ``"server_process_id"``, and
            ``"server_version"`` keys.
        """
        return self._base_service.server_info

    def _del_session(self):
        if self._session_instance:
            self._session_instance.delete()
        self._session_instance = None

    @property
    def session(self):
        """Allows to plan events call backs from the server:
        progress bar when workflows are running, logging...

        Returns
        -------
        ansys.dpf.core.session.Session
        """
        if not self._session_instance:
            from ansys.dpf.core import session

            self._session_instance = session.Session(self)
        return self._session_instance

    @property
    def _base_service(self):
        if not self._base_service_instance:
            from ansys.dpf.core.core import BaseService

            self._base_service_instance = BaseService(self, timeout=1)
        return self._base_service_instance

    @property
    @abc.abstractmethod
    def os(self):
        """Get the operating system of the server

        Returns
        -------
        os : str
            "nt" or "posix"
        """
        pass

    @property
    def on_docker(self):
        return self._docker_config.use_docker

    @property
    def docker_config(self):
        return self._docker_config

    @docker_config.setter
    def docker_config(self, val):
        self._docker_config = val

    @property
    @abc.abstractmethod
    def config(self):
        pass

    @abc.abstractmethod
    def shutdown(self):
        pass

    def release(self):
        """Clears the available Operators and Releases licenses when necessary.

        Notes
        -----
        Available with server's version starting at 6.0 (Ansys 2023R2).
        """
        self._base_service.release_dpf()

    def apply_context(self, context):
        """Defines the settings that will be used to load DPF's plugins.
        A DPF xml file can be used to list the plugins and set up variables.

        Parameters
        ----------
        context : ServerContext
            The context allows to choose which capabilities are available server side.

        Notes
        -----
        Available with server's version starting at 6.0 (Ansys 2023R2).
        """
        self._base_service.apply_context(context)
        self._context = context

    @property
    def context(self):
        """Returns the settings used to load DPF's plugins.
        To update the context server side, use
        :func:`ansys.dpf.core.BaseServer.server_types.apply_context`

        Returns
        -------
        ServerContext
        """
        return self._context

    def check_version(self, required_version, msg=None):
        """Check if the server version matches with a required version.

        Parameters
        ----------
        required_version : str
            Required version to compare with the server version.
        msg : str, optional
            Message for the raised exception if version requirements do not match.

        Raises
        ------
        dpf_errors : errors
            errors.DpfVersionNotSupported is raised if failure.

        Returns
        -------
        bool
            ``True`` if the server version meets the requirement.
        """
        from ansys.dpf.core.check_version import server_meet_version_and_raise

        return server_meet_version_and_raise(required_version, self, msg)

    def meet_version(self, required_version):
        """Check if the server version matches with a required version.

        Parameters
        ----------
        required_version : str
            Required version to compare with the server version.

        Returns
        -------
        bool
            ``True`` if the server version meets the requirement.
        """
        return server_meet_version(required_version, self)

    @property
    @abc.abstractmethod
    def local_server(self) -> bool:
        pass

    @local_server.setter
    @abc.abstractmethod
    def local_server(self, val):
        pass

    def __str__(self):
        return f"DPF Server: {self.info}"

    @abc.abstractmethod
    def __eq__(self, other_server):
        pass

    def __ne__(self, other_server):
        """Return true, if the servers are not equal"""
        return not self.__eq__(other_server)

    def __del__(self):
        try:
            if hasattr(core, "SERVER") and id(core.SERVER) == id(self):
                core.SERVER = None
        except:
            warnings.warn(traceback.format_exc())

        try:
            if hasattr(core, "_server_instances") and core._server_instances is not None:
                for i, server in enumerate(core._server_instances):
                    if server() == self:
                        core._server_instances.remove(server)
        except:
            warnings.warn(traceback.format_exc())


class CServer(BaseServer, ABC):
    """Abstract class for servers going through the DPFClientAPI"""

    def __init__(self, ansys_path=None, load_operators=True):
        super().__init__()
        self._own_process = False
        self.ansys_path = ansys_path
        self._client_api_path = load_api.load_client_api(ansys_path=ansys_path)

    @property
    def available_api_types(self):
        return "c_api"

    def get_api_for_type(self, capi, grpcapi):
        return capi

    def __del__(self):
        try:
            self._del_session()
            if self._own_process:
                self.shutdown()
            super().__del__()
        except:
            warnings.warn(traceback.format_exc())


class GrpcClient:
    def __init__(self, address=None):
        from ansys.dpf.gate import client_capi

        self._internal_obj = client_capi.ClientCAPI.client_new_full_address(address)
        client_capi.ClientCAPI.init_client_environment(self)

    def __del__(self):
        try:
            self._deleter_func[0](self._deleter_func[1](self))
        except:
            warnings.warn(traceback.format_exc())


class GrpcServer(CServer):
    """Server using the gRPC communication protocol"""

    def __init__(
        self,
        ansys_path=None,
        ip=LOCALHOST,
        port=DPF_DEFAULT_PORT,
        timeout=10,
        as_global=True,
        load_operators=True,
        launch_server=True,
        docker_config=RUNNING_DOCKER,
        use_pypim=True,
        num_connection_tryouts=3,
        context=server_context.SERVER_CONTEXT,
    ):
        # Load DPFClientAPI
        from ansys.dpf.core.misc import is_pypim_configured

        super().__init__(ansys_path=ansys_path, load_operators=load_operators)
        # Load Ans.Dpf.GrpcClient
        self._grpc_client_path = load_api.load_grpc_client(ansys_path=ansys_path)
        self._own_process = launch_server
        self._local_server = False

        address = f"{ip}:{port}"

        self._remote_instance = None
        if launch_server:
            if (
                is_pypim_configured()
                and not ansys_path
                and not docker_config.use_docker
                and use_pypim
            ):
                self._remote_instance = launch_remote_dpf()
                address = self._remote_instance.services["grpc"].uri
                ip = address.split(":")[-2]
                port = int(address.split(":")[-1])

            elif docker_config.use_docker:
                self.docker_config = server_factory.RunningDockerConfig(docker_config)
                launch_dpf_on_docker(
                    running_docker_config=self.docker_config,
                    ansys_path=ansys_path,
                    ip=ip,
                    port=port,
                    timeout=timeout,
                )
            else:
                launch_dpf(ansys_path, ip, port, timeout=timeout)
                self._local_server = True

        self._client = GrpcClient(address)

        # store port and ip for later reference
        self._address = address
        self._input_ip = ip
        self._input_port = port
        self.live = True
        self._create_shutdown_funcs()
        self._check_first_call(num_connection_tryouts)
        try:
            self._base_service.initialize_with_context(context)
            self._context = context
        except errors.DpfVersionNotSupported:
            pass
        self.set_as_global(as_global=as_global)

    def _check_first_call(self, num_connection_tryouts):
        for i in range(num_connection_tryouts):
            try:
                self.version
                break
            except errors.DPFServerException as e:
                if ("GOAWAY" not in str(e.args) and "unavailable" not in str(e.args)) or i == (
                    num_connection_tryouts - 1
                ):
                    raise e

    @property
    def version(self):
        from ansys.dpf.gate import data_processing_capi, integral_types

        api = data_processing_capi.DataProcessingCAPI
        major = integral_types.MutableInt32()
        minor = integral_types.MutableInt32()
        api.data_processing_get_server_version_on_client(self.client, major, minor)
        out = str(int(major)) + "." + str(int(minor))
        return out

    @property
    def os(self):
        from ansys.dpf.gate import data_processing_capi

        api = data_processing_capi.DataProcessingCAPI
        return api.data_processing_get_os_on_client(self.client)

    def _create_shutdown_funcs(self):
        from ansys.dpf.gate import data_processing_capi

        api = data_processing_capi.DataProcessingCAPI
        self._preparing_shutdown_func = (
            api.data_processing_prepare_shutdown,
            self.client,
        )
        self._shutdown_func = (api.data_processing_release_server, self.client)

    def shutdown(self):
        if self.live:
            if self._remote_instance:
                self._remote_instance.delete()
            try:
                if hasattr(self, "_preparing_shutdown_func"):
                    self._preparing_shutdown_func[0](self._preparing_shutdown_func[1])
            except Exception as e:
                warnings.warn("couldn't prepare shutdown: " + str(e.args))
            try:
                if hasattr(self, "_shutdown_func"):
                    self._shutdown_func[0](self._shutdown_func[1])
            except Exception as e:
                warnings.warn("couldn't shutdown server: " + str(e.args))

            self._docker_config.remove_docker_image()
            self.live = False

    def __eq__(self, other_server):
        """Return true, if ***** are equals"""
        if isinstance(other_server, GrpcServer):
            # """Return true, if the ip and the port are equals"""
            return self.address == other_server.address
        return False

    @property
    def client(self):
        return self._client

    @property
    def address(self):
        """Address of the server.

        Returns
        -------
        address : str
        """
        return self._address

    @property
    def ip(self):
        """IP address of the server.

        Returns
        -------
        ip : str
        """
        try:
            return self.info["server_ip"]
        except:
            return ""

    @property
    def port(self):
        """Port of the server.

        Returns
        -------
        port : int
        """
        try:
            return self.info["server_port"]
        except:
            return 0

    @property
    def external_ip(self):
        """Public IP address of the server.
        Is the same as  :func:`ansys.dpf.core.GrpcServer.ip` in all cases except
        for servers using a gateway:
        for example, servers running in Docker Images might have an internal
        :func:`ansys.dpf.core.GrpcServer.ip` different from the public
        :func:`ansys.dpf.core.GrpcServer.external_ip`, the latter should be used to get
        connected to the server from outside the Docker Image.

        Returns
        -------
        external_ip : str
        """
        return self._input_ip

    @property
    def external_port(self):
        """Public Port of the server.
        Is the same as  :func:`ansys.dpf.core.GrpcServer.port` in all cases except
        for servers using a gateway:
        for example, servers running in Docker Images might have an internal
        :func:`ansys.dpf.core.GrpcServer.port` different from the public
        :func:`ansys.dpf.core.GrpcServer.external_port`, the latter should be used to get
        connected to the server from outside the Docker Image.

        Returns
        -------
        port : int
        """
        return self._input_port

    @property
    def local_server(self):
        return self._local_server

    @local_server.setter
    def local_server(self, val):
        self._local_server = val

    @property
    def config(self):
        return server_factory.AvailableServerConfigs.GrpcServer


class InProcessServer(CServer):
    """Server using the InProcess communication protocol"""

    def __init__(
        self,
        ansys_path=None,
        as_global=True,
        load_operators=True,
        timeout=None,
        context=server_context.SERVER_CONTEXT,
    ):
        # Load DPFClientAPI
        super().__init__(ansys_path=ansys_path, load_operators=load_operators)
        # Load DataProcessingCore
        from ansys.dpf.gate.utils import data_processing_core_load_api

        name = "DataProcessingCore"
        path = _get_dll_path(name, ansys_path)
        try:
            data_processing_core_load_api(path, "common")
        except Exception as e:
            if not os.path.isdir(os.path.dirname(path)):
                raise NotADirectoryError(
                    f"DPF directory not found at {os.path.dirname(path)}"
                    f"Unable to locate the following file: {path}"
                )
            raise e
        try:
            self.apply_context(context)
        except errors.DpfVersionNotSupported:
            self._base_service.initialize_with_context(
                server_context.AvailableServerContexts.premium
            )
            self._context = server_context.AvailableServerContexts.premium
            pass
        self.set_as_global(as_global=as_global)

    @property
    def version(self):
        from ansys.dpf.gate import data_processing_capi, integral_types

        api = data_processing_capi.DataProcessingCAPI
        major = integral_types.MutableInt32()
        minor = integral_types.MutableInt32()
        api.data_processing_get_server_version(major, minor)
        out = str(int(major)) + "." + str(int(minor))
        return out

    @property
    def os(self):
        # Since it is InProcess, one could return the current os
        return os.name

    def shutdown(self):
        pass

    def __eq__(self, other_server):
        """Return true, if the ip and the port are equals"""
        return isinstance(other_server, InProcessServer)

    @property
    def client(self):
        return None

    @property
    def local_server(self):
        return True

    @local_server.setter
    def local_server(self, val):
        if not val:
            raise NotImplementedError("an in process server can only be local.")

    @property
    def config(self):
        return server_factory.AvailableServerConfigs.InProcessServer


class LegacyGrpcServer(BaseServer):
    """Provides an instance of the DPF server using InProcess gRPC.
    Kept for backward-compatibility with dpf servers <0.5.0.

    Parameters
    -----------
    ansys_path : str
        Path for the DPF executable.
    ip : str
        IP address of the remote or local instance to connect to. The
        default is ``"LOCALHOST"``.
    port : int
        Port to connect to the remote instance on. The default is
        ``"DPF_DEFAULT_PORT"``, which is 50054.
    timeout : float, optional
        Maximum number of seconds for the initialization attempt.
        The default is ``10``. Once the specified number of seconds
        passes, the connection fails.
    as_global : bool, optional
        Global variable that stores the IP address and port for the DPF
        module. All DPF objects created in this Python session will
        use this IP and port. The default is ``True``.
    load_operators : bool, optional
        Whether to automatically load the math operators. The default
        is ``True``.
    launch_server : bool, optional
        Whether to launch the server on Windows.
    docker_config : server_factory.DockerConfig, optional
        To start DPF server as a docker, specify the docker name here.
    use_pypim: bool, optional
        Whether to use PyPIM functionalities by default when a PyPIM environment is detected.
        Defaults to True.
    """

    def __init__(
        self,
        ansys_path=None,
        ip=LOCALHOST,
        port=DPF_DEFAULT_PORT,
        timeout=5,
        as_global=True,
        load_operators=True,
        launch_server=True,
        docker_config=RUNNING_DOCKER,
        use_pypim=True,
        context=server_context.SERVER_CONTEXT,
    ):
        """Start the DPF server."""
        # Use ansys.grpc.dpf
        from ansys.dpf.core.misc import is_pypim_configured

        super().__init__()

        self._info_instance = None
        self._own_process = launch_server
        self.live = False
        self._local_server = False

        # Load Ans.Dpf.Grpc?
        import grpc

        # check valid ip and port
        check_valid_ip(ip)
        if not isinstance(port, int):
            raise ValueError("Port must be an integer")

        address = f"{ip}:{port}"

        self._remote_instance = None
        if launch_server:
            if (
                is_pypim_configured()
                and not ansys_path
                and not docker_config.use_docker
                and use_pypim
            ):
                self._remote_instance = launch_remote_dpf()
                address = self._remote_instance.services["grpc"].uri
                ip = address.split(":")[-2]
                port = int(address.split(":")[-1])
            else:

                if docker_config.use_docker:
                    self.docker_config = server_factory.RunningDockerConfig(docker_config)
                    launch_dpf_on_docker(
                        running_docker_config=self.docker_config,
                        ansys_path=ansys_path,
                        ip=ip,
                        port=port,
                        timeout=timeout,
                    )
                else:
                    launch_dpf(ansys_path, ip, port, timeout=timeout)
                    self._local_server = True

        self.channel = grpc.insecure_channel(address)

        # store the address for later reference
        self._address = address
        self._input_ip = ip
        self._input_port = port
        self.live = True
        self.ansys_path = ansys_path
        self._stubs = {}

        self._create_shutdown_funcs()

        check_ansys_grpc_dpf_version(self, timeout)
        try:
            self._base_service.initialize_with_context(context)
            self._context = context
        except errors.DpfVersionNotSupported:
            pass
        self.set_as_global(as_global=as_global)

    def _create_shutdown_funcs(self):
        self._core_api = data_processing_grpcapi.DataProcessingGRPCAPI
        self._core_api.init_data_processing_environment(self)
        self._core_api.bind_delete_server_func(self)

    @property
    def client(self):
        return self

    @property
    def available_api_types(self):
        return list(self._stubs.values())

    def get_api_for_type(self, capi, grpcapi):
        return grpcapi

    def create_stub_if_necessary(self, stub_name, stub_type):
        if not (stub_name in self._stubs.keys()):
            self._stubs[stub_name] = stub_type(self.channel)

    def get_stub(self, stub_name):
        if not (stub_name in self._stubs.keys()):
            return None
        else:
            return self._stubs[stub_name]

    @property
    def ip(self):
        """IP address of the server.

        Returns
        -------
        ip : str
        """
        try:
            return self.info["server_ip"]
        except:
            return ""

    @property
    def port(self):
        """Port of the server.

        Returns
        -------
        port : int
        """
        try:
            return self.info["server_port"]
        except:
            return 0

    @property
    def external_ip(self):
        """Public IP address of the server.
        Is the same as  :func:`ansys.dpf.core.LegacyGrpcServer.ip` in all cases except
        for servers using a gateway:
        for example, servers running in Docker Images might have an internal
        :func:`ansys.dpf.core.LegacyGrpcServer.ip` different from the public
        :func:`ansys.dpf.core.LegacyGrpcServer.external_ip`, the latter should be used to get
        connected to the server from outside the Docker Image.

        Returns
        -------
        external_ip : str
        """
        return self._input_ip

    @property
    def external_port(self):
        """Public Port of the server.
        Is the same as  :func:`ansys.dpf.core.LegacyGrpcServer.port` in all cases except
        for servers using a gateway:
        for example, servers running in Docker Images might have an internal
        :func:`ansys.dpf.core.LegacyGrpcServer.port` different from the public
        :func:`ansys.dpf.core.LegacyGrpcServer.external_port`, the latter should be used to get
        connected to the server from outside the Docker Image.

        Returns
        -------
        port : int
        """
        return self._input_port

    @property
    def version(self):
        """Version of the server.

        Returns
        -------
        version : str
        """
        return self.info["server_version"]

    @property
    def os(self):
        """Get the operating system of the server

        Returns
        -------
        os : str
            "nt" or "posix"
        """
        return self.info["os"]

    @property
    def info(self):
        if not self._info_instance:
            self._info_instance = self._base_service.server_info
        return self._info_instance

    @property
    def local_server(self):
        return self._local_server

    @local_server.setter
    def local_server(self, val):
        self._local_server = val

    def shutdown(self):
        if self._own_process and self.live:
            try:
                if hasattr(self, "_preparing_shutdown_func"):
                    self._preparing_shutdown_func[0](self._preparing_shutdown_func[1])
            except Exception as e:
                warnings.warn("couldn't prepare shutdown: " + str(e.args))

            if self._remote_instance:
                self._remote_instance.delete()
            else:
                try:
                    if hasattr(self, "_shutdown_func"):
                        self._shutdown_func[0](self._shutdown_func[1])
                except Exception as e:
                    try:
                        if self.meet_version("4.0"):
                            warnings.warn(
                                f"couldn't properly release server: {str(e.args)}"
                                ".\n Killing process."
                            )
                        p = psutil.Process(self.info["server_process_id"])
                        p.kill()
                        time.sleep(0.01)
                    except:
                        warnings.warn(traceback.format_exc())

            self.live = False
            self._docker_config.remove_docker_image()

    @property
    def config(self):
        return server_factory.AvailableServerConfigs.LegacyGrpcServer

    def __eq__(self, other_server):
        """Return true, if the ip and the port are equals"""
        if isinstance(other_server, LegacyGrpcServer):
            return self.ip == other_server.ip and self.port == other_server.port
        return False

    def __del__(self):
        try:
            self._del_session()
            if self._own_process:
                self.shutdown()
            super().__del__()
        except:
            warnings.warn(traceback.format_exc())


# Python 3.10
# from typing import TypeAlias
# DpfServer: TypeAlias = LegacyGrpcServer
# Python <3.10
DpfServer = LegacyGrpcServer



"""
cyclic_analytic_usum_max
========================
Autogenerated DPF operator classes.
"""
from warnings import warn
from ansys.dpf.core.dpf_operator import Operator
from ansys.dpf.core.inputs import Input, _Inputs
from ansys.dpf.core.outputs import Output, _Outputs
from ansys.dpf.core.operators.specification import PinSpecification, Specification


class cyclic_analytic_usum_max(Operator):
    """Compute the maximum of the total deformation that can be expected on
    360 degrees

    Parameters
    ----------
    time_scoping : Scoping, optional
    mesh_scoping : ScopingsContainer or Scoping, optional
    fields_container : FieldsContainer
        Field container with the base and duplicate
        sectors
    bool_rotate_to_global : bool, optional
        Default is true
    cyclic_support : CyclicSupport


    Examples
    --------
    >>> from ansys.dpf import core as dpf

    >>> # Instantiate operator
    >>> op = dpf.operators.result.cyclic_analytic_usum_max()

    >>> # Make input connections
    >>> my_time_scoping = dpf.Scoping()
    >>> op.inputs.time_scoping.connect(my_time_scoping)
    >>> my_mesh_scoping = dpf.ScopingsContainer()
    >>> op.inputs.mesh_scoping.connect(my_mesh_scoping)
    >>> my_fields_container = dpf.FieldsContainer()
    >>> op.inputs.fields_container.connect(my_fields_container)
    >>> my_bool_rotate_to_global = bool()
    >>> op.inputs.bool_rotate_to_global.connect(my_bool_rotate_to_global)
    >>> my_cyclic_support = dpf.CyclicSupport()
    >>> op.inputs.cyclic_support.connect(my_cyclic_support)

    >>> # Instantiate operator and connect inputs in one line
    >>> op = dpf.operators.result.cyclic_analytic_usum_max(
    ...     time_scoping=my_time_scoping,
    ...     mesh_scoping=my_mesh_scoping,
    ...     fields_container=my_fields_container,
    ...     bool_rotate_to_global=my_bool_rotate_to_global,
    ...     cyclic_support=my_cyclic_support,
    ... )

    >>> # Get output data
    >>> result_fields_container = op.outputs.fields_container()
    """

    def __init__(
        self,
        time_scoping=None,
        mesh_scoping=None,
        fields_container=None,
        bool_rotate_to_global=None,
        cyclic_support=None,
        config=None,
        server=None,
    ):
        super().__init__(name="cyclic_analytic_usum_max", config=config, server=server)
        self._inputs = InputsCyclicAnalyticUsumMax(self)
        self._outputs = OutputsCyclicAnalyticUsumMax(self)
        if time_scoping is not None:
            self.inputs.time_scoping.connect(time_scoping)
        if mesh_scoping is not None:
            self.inputs.mesh_scoping.connect(mesh_scoping)
        if fields_container is not None:
            self.inputs.fields_container.connect(fields_container)
        if bool_rotate_to_global is not None:
            self.inputs.bool_rotate_to_global.connect(bool_rotate_to_global)
        if cyclic_support is not None:
            self.inputs.cyclic_support.connect(cyclic_support)

    @staticmethod
    def _spec():
        description = """Compute the maximum of the total deformation that can be expected on
            360 degrees"""
        spec = Specification(
            description=description,
            map_input_pin_spec={
                0: PinSpecification(
                    name="time_scoping",
                    type_names=["scoping", "vector<int32>"],
                    optional=True,
                    document="""""",
                ),
                1: PinSpecification(
                    name="mesh_scoping",
                    type_names=["scopings_container", "scoping", "vector<int32>"],
                    optional=True,
                    document="""""",
                ),
                2: PinSpecification(
                    name="fields_container",
                    type_names=["fields_container"],
                    optional=False,
                    document="""Field container with the base and duplicate
        sectors""",
                ),
                5: PinSpecification(
                    name="bool_rotate_to_global",
                    type_names=["bool"],
                    optional=True,
                    document="""Default is true""",
                ),
                16: PinSpecification(
                    name="cyclic_support",
                    type_names=["cyclic_support"],
                    optional=False,
                    document="""""",
                ),
            },
            map_output_pin_spec={
                0: PinSpecification(
                    name="fields_container",
                    type_names=["fields_container"],
                    optional=False,
                    document="""Fieldscontainer filled in""",
                ),
            },
        )
        return spec

    @staticmethod
    def default_config(server=None):
        """Returns the default config of the operator.

        This config can then be changed to the user needs and be used to
        instantiate the operator. The Configuration allows to customize
        how the operation will be processed by the operator.

        Parameters
        ----------
        server : server.DPFServer, optional
            Server with channel connected to the remote or local instance. When
            ``None``, attempts to use the global server.
        """
        return Operator.default_config(name="cyclic_analytic_usum_max", server=server)

    @property
    def inputs(self):
        """Enables to connect inputs to the operator

        Returns
        --------
        inputs : InputsCyclicAnalyticUsumMax
        """
        return super().inputs

    @property
    def outputs(self):
        """Enables to get outputs of the operator by evaluating it

        Returns
        --------
        outputs : OutputsCyclicAnalyticUsumMax
        """
        return super().outputs


class InputsCyclicAnalyticUsumMax(_Inputs):
    """Intermediate class used to connect user inputs to
    cyclic_analytic_usum_max operator.

    Examples
    --------
    >>> from ansys.dpf import core as dpf
    >>> op = dpf.operators.result.cyclic_analytic_usum_max()
    >>> my_time_scoping = dpf.Scoping()
    >>> op.inputs.time_scoping.connect(my_time_scoping)
    >>> my_mesh_scoping = dpf.ScopingsContainer()
    >>> op.inputs.mesh_scoping.connect(my_mesh_scoping)
    >>> my_fields_container = dpf.FieldsContainer()
    >>> op.inputs.fields_container.connect(my_fields_container)
    >>> my_bool_rotate_to_global = bool()
    >>> op.inputs.bool_rotate_to_global.connect(my_bool_rotate_to_global)
    >>> my_cyclic_support = dpf.CyclicSupport()
    >>> op.inputs.cyclic_support.connect(my_cyclic_support)
    """

    def __init__(self, op: Operator):
        super().__init__(cyclic_analytic_usum_max._spec().inputs, op)
        self._time_scoping = Input(
            cyclic_analytic_usum_max._spec().input_pin(0), 0, op, -1
        )
        self._inputs.append(self._time_scoping)
        self._mesh_scoping = Input(
            cyclic_analytic_usum_max._spec().input_pin(1), 1, op, -1
        )
        self._inputs.append(self._mesh_scoping)
        self._fields_container = Input(
            cyclic_analytic_usum_max._spec().input_pin(2), 2, op, -1
        )
        self._inputs.append(self._fields_container)
        self._bool_rotate_to_global = Input(
            cyclic_analytic_usum_max._spec().input_pin(5), 5, op, -1
        )
        self._inputs.append(self._bool_rotate_to_global)
        self._cyclic_support = Input(
            cyclic_analytic_usum_max._spec().input_pin(16), 16, op, -1
        )
        self._inputs.append(self._cyclic_support)

    @property
    def time_scoping(self):
        """Allows to connect time_scoping input to the operator.

        Parameters
        ----------
        my_time_scoping : Scoping

        Examples
        --------
        >>> from ansys.dpf import core as dpf
        >>> op = dpf.operators.result.cyclic_analytic_usum_max()
        >>> op.inputs.time_scoping.connect(my_time_scoping)
        >>> # or
        >>> op.inputs.time_scoping(my_time_scoping)
        """
        return self._time_scoping

    @property
    def mesh_scoping(self):
        """Allows to connect mesh_scoping input to the operator.

        Parameters
        ----------
        my_mesh_scoping : ScopingsContainer or Scoping

        Examples
        --------
        >>> from ansys.dpf import core as dpf
        >>> op = dpf.operators.result.cyclic_analytic_usum_max()
        >>> op.inputs.mesh_scoping.connect(my_mesh_scoping)
        >>> # or
        >>> op.inputs.mesh_scoping(my_mesh_scoping)
        """
        return self._mesh_scoping

    @property
    def fields_container(self):
        """Allows to connect fields_container input to the operator.

        Field container with the base and duplicate
        sectors

        Parameters
        ----------
        my_fields_container : FieldsContainer

        Examples
        --------
        >>> from ansys.dpf import core as dpf
        >>> op = dpf.operators.result.cyclic_analytic_usum_max()
        >>> op.inputs.fields_container.connect(my_fields_container)
        >>> # or
        >>> op.inputs.fields_container(my_fields_container)
        """
        return self._fields_container

    @property
    def bool_rotate_to_global(self):
        """Allows to connect bool_rotate_to_global input to the operator.

        Default is true

        Parameters
        ----------
        my_bool_rotate_to_global : bool

        Examples
        --------
        >>> from ansys.dpf import core as dpf
        >>> op = dpf.operators.result.cyclic_analytic_usum_max()
        >>> op.inputs.bool_rotate_to_global.connect(my_bool_rotate_to_global)
        >>> # or
        >>> op.inputs.bool_rotate_to_global(my_bool_rotate_to_global)
        """
        return self._bool_rotate_to_global

    @property
    def cyclic_support(self):
        """Allows to connect cyclic_support input to the operator.

        Parameters
        ----------
        my_cyclic_support : CyclicSupport

        Examples
        --------
        >>> from ansys.dpf import core as dpf
        >>> op = dpf.operators.result.cyclic_analytic_usum_max()
        >>> op.inputs.cyclic_support.connect(my_cyclic_support)
        >>> # or
        >>> op.inputs.cyclic_support(my_cyclic_support)
        """
        return self._cyclic_support


class OutputsCyclicAnalyticUsumMax(_Outputs):
    """Intermediate class used to get outputs from
    cyclic_analytic_usum_max operator.

    Examples
    --------
    >>> from ansys.dpf import core as dpf
    >>> op = dpf.operators.result.cyclic_analytic_usum_max()
    >>> # Connect inputs : op.inputs. ...
    >>> result_fields_container = op.outputs.fields_container()
    """

    def __init__(self, op: Operator):
        super().__init__(cyclic_analytic_usum_max._spec().outputs, op)
        self._fields_container = Output(
            cyclic_analytic_usum_max._spec().output_pin(0), 0, op
        )
        self._outputs.append(self._fields_container)

    @property
    def fields_container(self):
        """Allows to get fields_container output of the operator

        Returns
        ----------
        my_fields_container : FieldsContainer

        Examples
        --------
        >>> from ansys.dpf import core as dpf
        >>> op = dpf.operators.result.cyclic_analytic_usum_max()
        >>> # Connect inputs : op.inputs. ...
        >>> result_fields_container = op.outputs.fields_container()
        """  # noqa: E501
        return self._fields_container



"""
merge_fields
============
Autogenerated DPF operator classes.
"""
from warnings import warn
from ansys.dpf.core.dpf_operator import Operator
from ansys.dpf.core.inputs import Input, _Inputs
from ansys.dpf.core.outputs import Output, _Outputs
from ansys.dpf.core.operators.specification import PinSpecification, Specification


class merge_fields(Operator):
    """Assembles a set of fields into a unique one.

    Parameters
    ----------
    sum_merge : bool, optional
        Default is false. if true, redundant
        quantities are summed instead of
        being ignored.
    merged_support : AbstractFieldSupport, optional
        Already merged field support.
    fields1 : Field or FieldsContainer
        Either a fields container, a vector of fields
        to merge, or fields from pin 0 to ...
    fields2 : Field or FieldsContainer
        Either a fields container, a vector of fields
        to merge, or fields from pin 0 to ...


    Examples
    --------
    >>> from ansys.dpf import core as dpf

    >>> # Instantiate operator
    >>> op = dpf.operators.utility.merge_fields()

    >>> # Make input connections
    >>> my_sum_merge = bool()
    >>> op.inputs.sum_merge.connect(my_sum_merge)
    >>> my_merged_support = dpf.AbstractFieldSupport()
    >>> op.inputs.merged_support.connect(my_merged_support)
    >>> my_fields1 = dpf.Field()
    >>> op.inputs.fields1.connect(my_fields1)
    >>> my_fields2 = dpf.Field()
    >>> op.inputs.fields2.connect(my_fields2)

    >>> # Instantiate operator and connect inputs in one line
    >>> op = dpf.operators.utility.merge_fields(
    ...     sum_merge=my_sum_merge,
    ...     merged_support=my_merged_support,
    ...     fields1=my_fields1,
    ...     fields2=my_fields2,
    ... )

    >>> # Get output data
    >>> result_merged_field = op.outputs.merged_field()
    """

    def __init__(
        self,
        sum_merge=None,
        merged_support=None,
        fields1=None,
        fields2=None,
        config=None,
        server=None,
    ):
        super().__init__(name="merge::field", config=config, server=server)
        self._inputs = InputsMergeFields(self)
        self._outputs = OutputsMergeFields(self)
        if sum_merge is not None:
            self.inputs.sum_merge.connect(sum_merge)
        if merged_support is not None:
            self.inputs.merged_support.connect(merged_support)
        if fields1 is not None:
            self.inputs.fields1.connect(fields1)
        if fields2 is not None:
            self.inputs.fields2.connect(fields2)

    @staticmethod
    def _spec():
        description = """Assembles a set of fields into a unique one."""
        spec = Specification(
            description=description,
            map_input_pin_spec={
                -2: PinSpecification(
                    name="sum_merge",
                    type_names=["bool"],
                    optional=True,
                    document="""Default is false. if true, redundant
        quantities are summed instead of
        being ignored.""",
                ),
                -1: PinSpecification(
                    name="merged_support",
                    type_names=["abstract_field_support"],
                    optional=True,
                    document="""Already merged field support.""",
                ),
                0: PinSpecification(
                    name="fields",
                    type_names=["field", "fields_container"],
                    optional=False,
                    document="""Either a fields container, a vector of fields
        to merge, or fields from pin 0 to ...""",
                ),
                1: PinSpecification(
                    name="fields",
                    type_names=["field", "fields_container"],
                    optional=False,
                    document="""Either a fields container, a vector of fields
        to merge, or fields from pin 0 to ...""",
                ),
            },
            map_output_pin_spec={
                0: PinSpecification(
                    name="merged_field",
                    type_names=["field"],
                    optional=False,
                    document="""""",
                ),
            },
        )
        return spec

    @staticmethod
    def default_config(server=None):
        """Returns the default config of the operator.

        This config can then be changed to the user needs and be used to
        instantiate the operator. The Configuration allows to customize
        how the operation will be processed by the operator.

        Parameters
        ----------
        server : server.DPFServer, optional
            Server with channel connected to the remote or local instance. When
            ``None``, attempts to use the global server.
        """
        return Operator.default_config(name="merge::field", server=server)

    @property
    def inputs(self):
        """Enables to connect inputs to the operator

        Returns
        --------
        inputs : InputsMergeFields
        """
        return super().inputs

    @property
    def outputs(self):
        """Enables to get outputs of the operator by evaluating it

        Returns
        --------
        outputs : OutputsMergeFields
        """
        return super().outputs


class InputsMergeFields(_Inputs):
    """Intermediate class used to connect user inputs to
    merge_fields operator.

    Examples
    --------
    >>> from ansys.dpf import core as dpf
    >>> op = dpf.operators.utility.merge_fields()
    >>> my_sum_merge = bool()
    >>> op.inputs.sum_merge.connect(my_sum_merge)
    >>> my_merged_support = dpf.AbstractFieldSupport()
    >>> op.inputs.merged_support.connect(my_merged_support)
    >>> my_fields1 = dpf.Field()
    >>> op.inputs.fields1.connect(my_fields1)
    >>> my_fields2 = dpf.Field()
    >>> op.inputs.fields2.connect(my_fields2)
    """

    def __init__(self, op: Operator):
        super().__init__(merge_fields._spec().inputs, op)
        self._sum_merge = Input(merge_fields._spec().input_pin(-2), -2, op, -1)
        self._inputs.append(self._sum_merge)
        self._merged_support = Input(merge_fields._spec().input_pin(-1), -1, op, -1)
        self._inputs.append(self._merged_support)
        self._fields1 = Input(merge_fields._spec().input_pin(0), 0, op, 0)
        self._inputs.append(self._fields1)
        self._fields2 = Input(merge_fields._spec().input_pin(1), 1, op, 1)
        self._inputs.append(self._fields2)

    @property
    def sum_merge(self):
        """Allows to connect sum_merge input to the operator.

        Default is false. if true, redundant
        quantities are summed instead of
        being ignored.

        Parameters
        ----------
        my_sum_merge : bool

        Examples
        --------
        >>> from ansys.dpf import core as dpf
        >>> op = dpf.operators.utility.merge_fields()
        >>> op.inputs.sum_merge.connect(my_sum_merge)
        >>> # or
        >>> op.inputs.sum_merge(my_sum_merge)
        """
        return self._sum_merge

    @property
    def merged_support(self):
        """Allows to connect merged_support input to the operator.

        Already merged field support.

        Parameters
        ----------
        my_merged_support : AbstractFieldSupport

        Examples
        --------
        >>> from ansys.dpf import core as dpf
        >>> op = dpf.operators.utility.merge_fields()
        >>> op.inputs.merged_support.connect(my_merged_support)
        >>> # or
        >>> op.inputs.merged_support(my_merged_support)
        """
        return self._merged_support

    @property
    def fields1(self):
        """Allows to connect fields1 input to the operator.

        Either a fields container, a vector of fields
        to merge, or fields from pin 0 to ...

        Parameters
        ----------
        my_fields1 : Field or FieldsContainer

        Examples
        --------
        >>> from ansys.dpf import core as dpf
        >>> op = dpf.operators.utility.merge_fields()
        >>> op.inputs.fields1.connect(my_fields1)
        >>> # or
        >>> op.inputs.fields1(my_fields1)
        """
        return self._fields1

    @property
    def fields2(self):
        """Allows to connect fields2 input to the operator.

        Either a fields container, a vector of fields
        to merge, or fields from pin 0 to ...

        Parameters
        ----------
        my_fields2 : Field or FieldsContainer

        Examples
        --------
        >>> from ansys.dpf import core as dpf
        >>> op = dpf.operators.utility.merge_fields()
        >>> op.inputs.fields2.connect(my_fields2)
        >>> # or
        >>> op.inputs.fields2(my_fields2)
        """
        return self._fields2


class OutputsMergeFields(_Outputs):
    """Intermediate class used to get outputs from
    merge_fields operator.

    Examples
    --------
    >>> from ansys.dpf import core as dpf
    >>> op = dpf.operators.utility.merge_fields()
    >>> # Connect inputs : op.inputs. ...
    >>> result_merged_field = op.outputs.merged_field()
    """

    def __init__(self, op: Operator):
        super().__init__(merge_fields._spec().outputs, op)
        self._merged_field = Output(merge_fields._spec().output_pin(0), 0, op)
        self._outputs.append(self._merged_field)

    @property
    def merged_field(self):
        """Allows to get merged_field output of the operator

        Returns
        ----------
        my_merged_field : Field

        Examples
        --------
        >>> from ansys.dpf import core as dpf
        >>> op = dpf.operators.utility.merge_fields()
        >>> # Connect inputs : op.inputs. ...
        >>> result_merged_field = op.outputs.merged_field()
        """  # noqa: E501
        return self._merged_field



"""
elemental_difference
====================
Autogenerated DPF operator classes.
"""
from warnings import warn
from ansys.dpf.core.dpf_operator import Operator
from ansys.dpf.core.inputs import Input, _Inputs
from ansys.dpf.core.outputs import Output, _Outputs
from ansys.dpf.core.operators.specification import PinSpecification, Specification


class elemental_difference(Operator):
    """Transforms an Elemental Nodal or Nodal field into an Elemental field.
    Each elemental value is the maximum difference between the
    computed result for all nodes in this element. The result is
    computed on a given element scoping.

    Parameters
    ----------
    field : Field or FieldsContainer
        Field or fields container with only one field
        is expected
    mesh_scoping : Scoping, optional
        Average only on these entities
    mesh : MeshedRegion, optional
    through_layers : bool, optional
        The maximum elemental difference is taken
        through the different shell layers if
        true (default is false).


    Examples
    --------
    >>> from ansys.dpf import core as dpf

    >>> # Instantiate operator
    >>> op = dpf.operators.averaging.elemental_difference()

    >>> # Make input connections
    >>> my_field = dpf.Field()
    >>> op.inputs.field.connect(my_field)
    >>> my_mesh_scoping = dpf.Scoping()
    >>> op.inputs.mesh_scoping.connect(my_mesh_scoping)
    >>> my_mesh = dpf.MeshedRegion()
    >>> op.inputs.mesh.connect(my_mesh)
    >>> my_through_layers = bool()
    >>> op.inputs.through_layers.connect(my_through_layers)

    >>> # Instantiate operator and connect inputs in one line
    >>> op = dpf.operators.averaging.elemental_difference(
    ...     field=my_field,
    ...     mesh_scoping=my_mesh_scoping,
    ...     mesh=my_mesh,
    ...     through_layers=my_through_layers,
    ... )

    >>> # Get output data
    >>> result_field = op.outputs.field()
    """

    def __init__(
        self,
        field=None,
        mesh_scoping=None,
        mesh=None,
        through_layers=None,
        config=None,
        server=None,
    ):
        super().__init__(name="elemental_difference", config=config, server=server)
        self._inputs = InputsElementalDifference(self)
        self._outputs = OutputsElementalDifference(self)
        if field is not None:
            self.inputs.field.connect(field)
        if mesh_scoping is not None:
            self.inputs.mesh_scoping.connect(mesh_scoping)
        if mesh is not None:
            self.inputs.mesh.connect(mesh)
        if through_layers is not None:
            self.inputs.through_layers.connect(through_layers)

    @staticmethod
    def _spec():
        description = """Transforms an Elemental Nodal or Nodal field into an Elemental field.
            Each elemental value is the maximum difference between the
            computed result for all nodes in this element. The result
            is computed on a given element scoping."""
        spec = Specification(
            description=description,
            map_input_pin_spec={
                0: PinSpecification(
                    name="field",
                    type_names=["field", "fields_container"],
                    optional=False,
                    document="""Field or fields container with only one field
        is expected""",
                ),
                1: PinSpecification(
                    name="mesh_scoping",
                    type_names=["scoping"],
                    optional=True,
                    document="""Average only on these entities""",
                ),
                7: PinSpecification(
                    name="mesh",
                    type_names=["abstract_meshed_region"],
                    optional=True,
                    document="""""",
                ),
                10: PinSpecification(
                    name="through_layers",
                    type_names=["bool"],
                    optional=True,
                    document="""The maximum elemental difference is taken
        through the different shell layers if
        true (default is false).""",
                ),
            },
            map_output_pin_spec={
                0: PinSpecification(
                    name="field",
                    type_names=["field"],
                    optional=False,
                    document="""""",
                ),
            },
        )
        return spec

    @staticmethod
    def default_config(server=None):
        """Returns the default config of the operator.

        This config can then be changed to the user needs and be used to
        instantiate the operator. The Configuration allows to customize
        how the operation will be processed by the operator.

        Parameters
        ----------
        server : server.DPFServer, optional
            Server with channel connected to the remote or local instance. When
            ``None``, attempts to use the global server.
        """
        return Operator.default_config(name="elemental_difference", server=server)

    @property
    def inputs(self):
        """Enables to connect inputs to the operator

        Returns
        --------
        inputs : InputsElementalDifference
        """
        return super().inputs

    @property
    def outputs(self):
        """Enables to get outputs of the operator by evaluating it

        Returns
        --------
        outputs : OutputsElementalDifference
        """
        return super().outputs


class InputsElementalDifference(_Inputs):
    """Intermediate class used to connect user inputs to
    elemental_difference operator.

    Examples
    --------
    >>> from ansys.dpf import core as dpf
    >>> op = dpf.operators.averaging.elemental_difference()
    >>> my_field = dpf.Field()
    >>> op.inputs.field.connect(my_field)
    >>> my_mesh_scoping = dpf.Scoping()
    >>> op.inputs.mesh_scoping.connect(my_mesh_scoping)
    >>> my_mesh = dpf.MeshedRegion()
    >>> op.inputs.mesh.connect(my_mesh)
    >>> my_through_layers = bool()
    >>> op.inputs.through_layers.connect(my_through_layers)
    """

    def __init__(self, op: Operator):
        super().__init__(elemental_difference._spec().inputs, op)
        self._field = Input(elemental_difference._spec().input_pin(0), 0, op, -1)
        self._inputs.append(self._field)
        self._mesh_scoping = Input(elemental_difference._spec().input_pin(1), 1, op, -1)
        self._inputs.append(self._mesh_scoping)
        self._mesh = Input(elemental_difference._spec().input_pin(7), 7, op, -1)
        self._inputs.append(self._mesh)
        self._through_layers = Input(
            elemental_difference._spec().input_pin(10), 10, op, -1
        )
        self._inputs.append(self._through_layers)

    @property
    def field(self):
        """Allows to connect field input to the operator.

        Field or fields container with only one field
        is expected

        Parameters
        ----------
        my_field : Field or FieldsContainer

        Examples
        --------
        >>> from ansys.dpf import core as dpf
        >>> op = dpf.operators.averaging.elemental_difference()
        >>> op.inputs.field.connect(my_field)
        >>> # or
        >>> op.inputs.field(my_field)
        """
        return self._field

    @property
    def mesh_scoping(self):
        """Allows to connect mesh_scoping input to the operator.

        Average only on these entities

        Parameters
        ----------
        my_mesh_scoping : Scoping

        Examples
        --------
        >>> from ansys.dpf import core as dpf
        >>> op = dpf.operators.averaging.elemental_difference()
        >>> op.inputs.mesh_scoping.connect(my_mesh_scoping)
        >>> # or
        >>> op.inputs.mesh_scoping(my_mesh_scoping)
        """
        return self._mesh_scoping

    @property
    def mesh(self):
        """Allows to connect mesh input to the operator.

        Parameters
        ----------
        my_mesh : MeshedRegion

        Examples
        --------
        >>> from ansys.dpf import core as dpf
        >>> op = dpf.operators.averaging.elemental_difference()
        >>> op.inputs.mesh.connect(my_mesh)
        >>> # or
        >>> op.inputs.mesh(my_mesh)
        """
        return self._mesh

    @property
    def through_layers(self):
        """Allows to connect through_layers input to the operator.

        The maximum elemental difference is taken
        through the different shell layers if
        true (default is false).

        Parameters
        ----------
        my_through_layers : bool

        Examples
        --------
        >>> from ansys.dpf import core as dpf
        >>> op = dpf.operators.averaging.elemental_difference()
        >>> op.inputs.through_layers.connect(my_through_layers)
        >>> # or
        >>> op.inputs.through_layers(my_through_layers)
        """
        return self._through_layers


class OutputsElementalDifference(_Outputs):
    """Intermediate class used to get outputs from
    elemental_difference operator.

    Examples
    --------
    >>> from ansys.dpf import core as dpf
    >>> op = dpf.operators.averaging.elemental_difference()
    >>> # Connect inputs : op.inputs. ...
    >>> result_field = op.outputs.field()
    """

    def __init__(self, op: Operator):
        super().__init__(elemental_difference._spec().outputs, op)
        self._field = Output(elemental_difference._spec().output_pin(0), 0, op)
        self._outputs.append(self._field)

    @property
    def field(self):
        """Allows to get field output of the operator

        Returns
        ----------
        my_field : Field

        Examples
        --------
        >>> from ansys.dpf import core as dpf
        >>> op = dpf.operators.averaging.elemental_difference()
        >>> # Connect inputs : op.inputs. ...
        >>> result_field = op.outputs.field()
        """  # noqa: E501
        return self._field



from ansys.dpf.core.operator_specification import SpecificationBase, PinSpecification


class Specification(SpecificationBase):
    def __init__(self, description=None, map_input_pin_spec=None, map_output_pin_spec=None, config=None):

        self._description = description
        self._map_input_pin_spec = map_input_pin_spec
        self._map_output_pin_spec = map_output_pin_spec



    @property
    def map_input_pin_spec(self):
        return self._map_input_pin_spec

    @property
    def map_output_pin_spec(self):
        return self._map_output_pin_spec

    @property
    def inputs(self):
        return self._map_input_pin_spec

    @property
    def outputs(self):
        return self._map_output_pin_spec

    def output_pin(self, pin_num):
        return self._map_output_pin_spec[pin_num]

    def input_pin(self, pin_num):
        return self._map_input_pin_spec[pin_num]

    @property
    def description(self):
        return self._description

    def __str__(self):
        out = ""
        for key, i in self._asdict().items():
            out += key + ": " + str(i) + "\n\n"
        return out



"""
import_symbolic_workflow
========================
Autogenerated DPF operator classes.
"""
from warnings import warn
from ansys.dpf.core.dpf_operator import Operator
from ansys.dpf.core.inputs import Input, _Inputs
from ansys.dpf.core.outputs import Output, _Outputs
from ansys.dpf.core.operators.specification import PinSpecification, Specification


class import_symbolic_workflow(Operator):
    """Reads a file or string holding a Symbolic Workflow and instantiate a
    WorkFlow with its data.

    Parameters
    ----------
    string_or_path : str or DataSources
    format : int, optional
        0 is ascii format and 1 is binary, default is
        0.


    Examples
    --------
    >>> from ansys.dpf import core as dpf

    >>> # Instantiate operator
    >>> op = dpf.operators.serialization.import_symbolic_workflow()

    >>> # Make input connections
    >>> my_string_or_path = str()
    >>> op.inputs.string_or_path.connect(my_string_or_path)
    >>> my_format = int()
    >>> op.inputs.format.connect(my_format)

    >>> # Instantiate operator and connect inputs in one line
    >>> op = dpf.operators.serialization.import_symbolic_workflow(
    ...     string_or_path=my_string_or_path,
    ...     format=my_format,
    ... )

    >>> # Get output data
    >>> result_workflow = op.outputs.workflow()
    """

    def __init__(self, string_or_path=None, format=None, config=None, server=None):
        super().__init__(name="import_symbolic_workflow", config=config, server=server)
        self._inputs = InputsImportSymbolicWorkflow(self)
        self._outputs = OutputsImportSymbolicWorkflow(self)
        if string_or_path is not None:
            self.inputs.string_or_path.connect(string_or_path)
        if format is not None:
            self.inputs.format.connect(format)

    @staticmethod
    def _spec():
        description = """Reads a file or string holding a Symbolic Workflow and instantiate a
            WorkFlow with its data."""
        spec = Specification(
            description=description,
            map_input_pin_spec={
                0: PinSpecification(
                    name="string_or_path",
                    type_names=["string", "data_sources"],
                    optional=False,
                    document="""""",
                ),
                2: PinSpecification(
                    name="format",
                    type_names=["int32"],
                    optional=True,
                    document="""0 is ascii format and 1 is binary, default is
        0.""",
                ),
            },
            map_output_pin_spec={
                0: PinSpecification(
                    name="workflow",
                    type_names=["workflow"],
                    optional=False,
                    document="""""",
                ),
            },
        )
        return spec

    @staticmethod
    def default_config(server=None):
        """Returns the default config of the operator.

        This config can then be changed to the user needs and be used to
        instantiate the operator. The Configuration allows to customize
        how the operation will be processed by the operator.

        Parameters
        ----------
        server : server.DPFServer, optional
            Server with channel connected to the remote or local instance. When
            ``None``, attempts to use the global server.
        """
        return Operator.default_config(name="import_symbolic_workflow", server=server)

    @property
    def inputs(self):
        """Enables to connect inputs to the operator

        Returns
        --------
        inputs : InputsImportSymbolicWorkflow
        """
        return super().inputs

    @property
    def outputs(self):
        """Enables to get outputs of the operator by evaluating it

        Returns
        --------
        outputs : OutputsImportSymbolicWorkflow
        """
        return super().outputs


class InputsImportSymbolicWorkflow(_Inputs):
    """Intermediate class used to connect user inputs to
    import_symbolic_workflow operator.

    Examples
    --------
    >>> from ansys.dpf import core as dpf
    >>> op = dpf.operators.serialization.import_symbolic_workflow()
    >>> my_string_or_path = str()
    >>> op.inputs.string_or_path.connect(my_string_or_path)
    >>> my_format = int()
    >>> op.inputs.format.connect(my_format)
    """

    def __init__(self, op: Operator):
        super().__init__(import_symbolic_workflow._spec().inputs, op)
        self._string_or_path = Input(
            import_symbolic_workflow._spec().input_pin(0), 0, op, -1
        )
        self._inputs.append(self._string_or_path)
        self._format = Input(import_symbolic_workflow._spec().input_pin(2), 2, op, -1)
        self._inputs.append(self._format)

    @property
    def string_or_path(self):
        """Allows to connect string_or_path input to the operator.

        Parameters
        ----------
        my_string_or_path : str or DataSources

        Examples
        --------
        >>> from ansys.dpf import core as dpf
        >>> op = dpf.operators.serialization.import_symbolic_workflow()
        >>> op.inputs.string_or_path.connect(my_string_or_path)
        >>> # or
        >>> op.inputs.string_or_path(my_string_or_path)
        """
        return self._string_or_path

    @property
    def format(self):
        """Allows to connect format input to the operator.

        0 is ascii format and 1 is binary, default is
        0.

        Parameters
        ----------
        my_format : int

        Examples
        --------
        >>> from ansys.dpf import core as dpf
        >>> op = dpf.operators.serialization.import_symbolic_workflow()
        >>> op.inputs.format.connect(my_format)
        >>> # or
        >>> op.inputs.format(my_format)
        """
        return self._format


class OutputsImportSymbolicWorkflow(_Outputs):
    """Intermediate class used to get outputs from
    import_symbolic_workflow operator.

    Examples
    --------
    >>> from ansys.dpf import core as dpf
    >>> op = dpf.operators.serialization.import_symbolic_workflow()
    >>> # Connect inputs : op.inputs. ...
    >>> result_workflow = op.outputs.workflow()
    """

    def __init__(self, op: Operator):
        super().__init__(import_symbolic_workflow._spec().outputs, op)
        self._workflow = Output(import_symbolic_workflow._spec().output_pin(0), 0, op)
        self._outputs.append(self._workflow)

    @property
    def workflow(self):
        """Allows to get workflow output of the operator

        Returns
        ----------
        my_workflow : Workflow

        Examples
        --------
        >>> from ansys.dpf import core as dpf
        >>> op = dpf.operators.serialization.import_symbolic_workflow()
        >>> # Connect inputs : op.inputs. ...
        >>> result_workflow = op.outputs.workflow()
        """  # noqa: E501
        return self._workflow



"""
.. _ref_faces_apis:

Faces
=====
"""
import numpy as np
from ansys.dpf.core import scoping
from ansys.dpf.core.common import face_properties
from ansys.dpf.core.elements import element_types
from ansys.dpf.core.check_version import version_requires


@version_requires("7.0")
class Face:
    """
    Contains all properties of a face of a mesh.

    The face is created from the
    :class:`MeshedRegion <ansys.dpf.core.meshed_region.MeshedRegion>` class.
    Properties include the face ID, index, type, shape, and connectivity.

    Parameters
    ----------
    mesh : :class:`ansys.dpf.core.meshed_region.MeshedRegion`
        Mesh containing the face.
    faceid : int
        Number (ID) of the face.
    index : int
        Fortran-based (1-based) index of the face in the result.
    nodes : list
        List of DPF nodes belonging to the face.

    Examples
    --------
    Extract a single face from a meshed region.

    >>> import ansys.dpf.core as dpf
    >>> from ansys.dpf.core import examples
    >>> model = dpf.Model(examples.fluid_axial_model())
    >>> faces = model.metadata.meshed_region.faces
    >>> face = faces[0]

    List the coordinates belonging to the first node of the face.

    >>> face.nodes[0].coordinates
    [-0.030426240620025163, -0.05908951107677226, -0.034248966723680496]

    Notes
    -----
    Class available with server's version starting at 7.0 (2024 R1 pre0).
    """

    def __init__(self, mesh, faceid, index, nodes):
        self._id = faceid
        self._index = index
        self._nodes = nodes
        self._mesh = mesh

    @property
    def node_ids(self):
        """
        IDs of all nodes in the face.

        Returns
        --------
        list
           List of IDs for all nodes in the face.

        Examples
        --------
        >>> import ansys.dpf.core as dpf
        >>> from ansys.dpf.core import examples
        >>> model = dpf.Model(examples.fluid_axial_model())
        >>> faces = model.metadata.meshed_region.faces
        >>> face = faces[0]
        >>> face.node_ids
        [11291, 11416, 11455, 11325]

        """
        return [node.id for node in self._nodes]

    @property
    def id(self) -> int:
        """
        ID of the face.

        Returns
        -------
        int
            ID of the face.

        """
        return self._id

    @property
    def index(self) -> int:
        """
        Index of the face in the result.

        Returns
        -------
        int
            Index of the face in the result. This uses zero-based indexing starting at ``0``.

        """
        return self._index

    @property
    def nodes(self):
        """
        All nodes in the face.
        Returns
        --------
        list
            List of all nodes in the face.

        Examples
        --------
        >>> import ansys.dpf.core as dpf
        >>> from ansys.dpf.core import examples
        >>> model = dpf.Model(examples.fluid_axial_model())
        >>> faces = model.metadata.meshed_region.faces
        >>> face = faces[0]
        >>> first_node = face.nodes[0]

        """
        return self._nodes

    @property
    def n_nodes(self) -> int:
        """
        Number of nodes in the face.

        Returns
        -------
        int
            Number of nodes.

        """
        return len(self._nodes)

    def __str__(self):
        txt = "DPF Face %d\n" % self.id
        txt += "\tIndex:      %7d\n" % self.index
        txt += "\tNodes:      %7d\n" % self.n_nodes
        txt += f"\tType:       {self.type}\n"
        return txt

    @property
    def type(self) -> int:
        """
        Type of the face.

        Returns
        -------
        int
            Type of the face. For more information, see
            :class:`ansys.dpf.core.elements.element_types`.

        Examples
        --------
        >>> import ansys.dpf.core as dpf
        >>> from ansys.dpf.core import examples
        >>> model = dpf.Model(examples.fluid_axial_model())
        >>> faces = model.metadata.meshed_region.faces
        >>> face = faces[0]
        >>> face.type
        <element_types.Quad4: 16>

        """
        return self._get_type()

    def _get_type(self):
        """Retrieve the Ansys element type."""
        return element_types(
            self._mesh.property_field(face_properties.faces_type).get_entity_data(self._index)[0]
        )

    @property
    def connectivity(self):
        """
        Ordered list of node indices of the face.

        Returns
        --------
        list
            Ordered list of node indices.

        """
        list = []
        for node in self._nodes:
            list.append(node.index)
        return list


@version_requires("7.0")
class Faces:
    """
    Contains faces belonging to a meshed region.

    Parameters
    ----------
    mesh : str
        Name of the meshed region.

    Examples
    --------
    >>> import ansys.dpf.core as dpf
    >>> from ansys.dpf.core import examples
    >>> model = dpf.Model(examples.fluid_axial_model())
    >>> faces = model.metadata.meshed_region.faces
    >>> faces.n_faces
    44242

    Notes
    -----
    Class available with server's version starting at 7.0 (2024 R1 pre0).
    """

    def __init__(self, mesh):
        self._mesh = mesh
        self._server = mesh._server
        self._mapping_id_to_index = None

    def __str__(self):
        return "DPF Faces object with %d faces" % len(self)

    def __getitem__(self, index):
        """Retrieves face based on an index."""
        return self.face_by_index(index)

    def __len__(self):
        return self.n_faces

    def __iter__(self):
        for i in range(len(self)):
            yield self[i]

    def face_by_id(self, id) -> Face:
        """
        Retrieve a face by face ID.

        Parameters
        ----------
        id : int
            Number (ID) of the face.

        Returns
        -------
        Face
            Face object.

        """
        return self.__get_face(faceid=id)

    def face_by_index(self, index) -> Face:
        """
        Retrieve a face using its index.

        Parameters
        ----------
        index : int
            Zero-based index.

        Returns
        -------
        Face
            Yield face.

        Examples
        --------
        faces.face_by_index(0)

        Notes
        -----
        This is equivalent to ``faces[0]``

        """
        return self.__get_face(faceindex=index)

    def __get_face(self, faceindex=None, faceid=None):
        """
        Retrieve the face by ID or index.

        Parameters
        ----------
        faceid : int, optional
            ID of the face. The default is ``None``.
        faceindex : int, optional
            Index of the face. The default is ``None``.

        Returns
        -------
        face : Face
        """
        if faceindex is None:
            faceindex = self._mesh.property_field(face_properties.faces_type).scoping.index(faceid)
        elif faceid is None:
            faceid = self._mesh.property_field(face_properties.faces_type).scoping.id(faceindex)
        nodeIdx = self._mesh.property_field(
            face_properties.faces_nodes_connectivity
        ).get_entity_data(faceindex)
        nodesOut = [self._mesh.nodes.node_by_index(node_index) for node_index in nodeIdx]
        return Face(self._mesh, faceid, faceindex, nodesOut)

    @property
    def scoping(self) -> scoping.Scoping:
        """
        Scoping of the faces.

        Returns
        -------
        :class:`ansys.dpf.core.scoping.Scoping`

        Examples
        --------
        >>> import ansys.dpf.core as dpf
        >>> from ansys.dpf.core import examples
        >>> model = dpf.Model(examples.fluid_axial_model())
        >>> faces = model.metadata.meshed_region.faces
        >>> my_scoping = faces.scoping

        """
        return self._mesh.property_field(face_properties.faces_type).scoping

    @property
    def faces_type_field(self):
        """
        Field of all faces types.

        Returns
        -------
        :class:`ansys.dpf.core.field.Field`
            Field of all faces types.

        Examples
        --------
        >>> import ansys.dpf.core as dpf
        >>> from ansys.dpf.core import examples
        >>> model = dpf.Model(examples.fluid_axial_model())
        >>> faces = model.metadata.meshed_region.faces
        >>> field = faces.faces_type_field
        >>> print(field.data)
        [16 16 16 ... 16 16 16]

        """
        return self._mesh.field_of_properties(face_properties.faces_type)

    @property
    def faces_nodes_connectivity_field(self):
        """
        Field containing for each face ID the node indices connected to the face.

        Returns
        -------
        :class:`ansys.dpf.core.field.Field`
            Field containing for each face ID the node indices connected to the face.

        Examples
        --------
        >>> import ansys.dpf.core as dpf
        >>> from ansys.dpf.core import examples
        >>> model = dpf.Model(examples.fluid_axial_model())
        >>> faces = model.metadata.meshed_region.faces
        >>> field = faces.faces_nodes_connectivity_field
        >>> field.get_entity_data(1)
        DPFArray([11415, 11347, 11387, 11454])

        """
        return self._mesh.property_field(face_properties.faces_nodes_connectivity)

    @property
    def n_faces(self) -> int:
        """Number of faces"""
        return self._mesh._api.meshed_region_get_num_faces(self._mesh)

    def _build_mapping_id_to_index(self):
        """Retrieve the mapping between the IDs and indices of the entity."""
        return {eid: i for i, eid in enumerate(self.scoping.ids)}

    @property
    def mapping_id_to_index(self) -> dict:
        """
        Mapping between the IDs and indices of the entity.

        This property is useful for mapping scalar results from a field to the meshed region.

        Examples
        --------
        >>> import ansys.dpf.core as dpf
        >>> from ansys.dpf.core import examples
        >>> model = dpf.Model(examples.fluid_axial_model())
        >>> meshed_region = model.metadata.meshed_region
        >>> map = meshed_region.faces.mapping_id_to_index

        """
        if self._mapping_id_to_index is None:
            self._mapping_id_to_index = self._build_mapping_id_to_index()
        return self._mapping_id_to_index

    def map_scoping(self, external_scope):
        """
        Retrieve the indices to map the scoping of these faces to
        the scoping of a field.

        Parameters
        ----------
        external_scope : :class:`ansys.dpf.core.scoping.Scoping`
            Scoping to map to.

        Returns
        -------
        indices : numpy.ndarray
            List of indices to map from the external scope to the
            scoping of these faces.
        mask : numpy.ndarray
            Members of the external scope that are in the face scoping.

        Examples
        --------
        Return the indices that map a field to a faces collection.

        >>> import ansys.dpf.core as dpf
        >>> from ansys.dpf.core import examples
        >>> model = dpf.Model(examples.fluid_axial_model())
        >>> faces = model.metadata.meshed_region.faces
        >>> m = model.results.mass_flow_rate()
        >>> field = m.outputs.fields_container()[0]
        >>> ind, mask = faces.map_scoping(field.scoping)

        """
        if external_scope.location in ["Nodal", "NodalElemental", "Elemental", "ElementalNodal"]:
            raise ValueError('Input scope location must be "Faces"')
        arr = np.array(list(map(self.mapping_id_to_index.get, external_scope.ids)))
        mask = arr != None
        ind = arr[mask].astype(np.int32)
        return ind, mask
